#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.soso import *
from .shdr import *
from supyr_struct.defs.tag_def import TagDef

bump_properties = Struct("bump_properties",
    Float("bump_scale"),
    dependency_stubbs("bump_map", "bitm"),
    )

soso_attrs = Struct("soso_attrs",
    #Model Shader Properties
    model_shader,

    Pad(16),
    #Color-Change
    SEnum16("color_change_source", *function_names),

    Pad(30),
    #Self-Illumination
    self_illumination,

    Pad(12),
    #Diffuse, Multipurpose, and Detail Maps
    maps,

    # this padding is the reflexive for the OS shader model extension
    Pad(12),

    #Texture Scrolling Animation
    texture_scrolling,

    Pad(8),
    #Reflection Properties
    reflection,

    Pad(16),
    bump_properties,
    SIZE=440
    )

soso_body = Struct("tagdata",
    shdr_attrs,
    soso_attrs
    )

def get():
    return soso_def

soso_def = TagDef("soso",
    blam_header_stubbs('soso', 3),
    soso_body,

    ext=".shader_model", endian=">",  # increment to differentiate it from halo soso
    tag_cls=ShdrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .coll import *
from .objs.sbsp import SbspTag
from supyr_struct.defs.block_def import BlockDef

cluster_fog_tooltip = (
    "Unknown flag is set if negative.\n"
    "Add 0x8000 to get fog index."
    )


# the order is an array of vertices first, then an array of lightmap vertices.
#
uncompressed_vertex = QStruct("uncompressed_vertex",
    Float('position_x'), Float('position_y'), Float('position_z'),
    Float('normal_i'),   Float('normal_j'),   Float('normal_k'),
    Float('binormal_i'), Float('binormal_j'), Float('binormal_k'),
    Float('tangent_i'),  Float('tangent_j'),  Float('tangent_k'),

    Float('tex_coord_u'), Float('tex_coord_v'),
    SIZE=56
    )

compressed_vertex = QStruct("compressed_vertex",
    Float('position_x'), Float('position_y'), Float('position_z'),
    UInt32('normal'),
    UInt32('binormal'),
    UInt32('tangent'),

    Float('tex_coord_u'), Float('tex_coord_v'),
    SIZE=32
    )

uncompressed_lightmap_vertex = QStruct("uncompressed_lightmap_vertex",
    # this normal is the direction the light is hitting from, and
    # is used for calculating dynamic shadows on dynamic objects
    Float('normal_i'), Float('normal_j'), Float('normal_k'),
    Float('u'), Float('v'),
    SIZE=20
    )

compressed_lightmap_vertex = QStruct("compressed_lightmap_vertex",
    # this normal is the direction the light is hitting from, and
    # is used for calculating dynamic shadows on dynamic objects
    UInt32('normal'),
    SInt16('u', UNIT_SCALE=1/32767, MIN=-32767, WIDGET_WIDTH=10),
    SInt16('v', UNIT_SCALE=1/32767, MIN=-32767, WIDGET_WIDTH=10),
    SIZE=8
    )

plane = QStruct("plane",
    Float("i", MIN=-1.0, MAX=1.0),
    Float("j", MIN=-1.0, MAX=1.0),
    Float("k", MIN=-1.0, MAX=1.0),
    Float("d"),
    SIZE=16, ORIENT='h'
    )

vertex = QStruct("vertex", INCLUDE=xyz_float, SIZE=12)

collision_material = Struct("collision_material",
    dependency("shader", valid_shaders),
    FlUInt32("unknown", VISIBLE=False),
    SIZE=20
    )

collision_bsp = Struct("collision_bsp", INCLUDE=permutation_bsp)
fast_collision_bsp = Struct("collision_bsp", INCLUDE=fast_permutation_bsp)

node = Struct("node",
    # these dont get byteswapped going from meta to tag
    BytesRaw("unknown", SIZE=6),
    #QStruct("unknown_0", UInt8("val0"), SInt8("val1"), ORIENT="h"),
    #QStruct("unknown_1", UInt8("val0"), SInt8("val1"), ORIENT="h"),
    #QStruct("unknown_2", UInt8("val0"), SInt8("val1"), ORIENT="h"),
    SIZE=6
    )

leaf = QStruct("leaf",
    # these unknowns are in the tag and are preserved in the meta
    FlSInt16("unknown0", VISIBLE=False),
    FlSInt16("unknown1", VISIBLE=False),
    FlSInt16("unknown2", VISIBLE=False),
    FlSInt16("unknown3", VISIBLE=False),

    SInt16("cluster"),
    SInt16("surface_reference_count"),
    SInt32("surface_references"),
    SIZE=16,
    )

leaf_surface = QStruct("leaf_surface",
    SInt32("surface"),
    SInt32("node"),
    SIZE=8, ORIENT='h'
    )

surface = QStruct("surface",
    SInt16("a"),
    SInt16("b"),
    SInt16("c"),
    SIZE=6, ORIENT='h', COMMENT="""
This is a renderable surface(visible geometry and lightmap geometry)"""
    )

material = Struct("material",
    dependency("shader", valid_shaders),
    SInt16("shader_permutation"),
    Bool16("flags",
        "coplanar",
        "fog_plane",
        ),
    SInt32("surfaces", EDITABLE=False,
        TOOLTIP=("The offset into the surfaces array that this\n"
                 "lightmap materials surfaces are located at.")
        ),
    SInt32("surface_count", EDITABLE=False,
        TOOLTIP=("The number of surfaces in the array belonging\n"
                 "to this lightmap material.")
        ),
    QStruct("centroid", INCLUDE=xyz_float),
    QStruct("ambient_color", INCLUDE=rgb_float),
    SInt16("distant_light_count"),
    Pad(2),

    QStruct("distant_light_0_color", INCLUDE=rgb_float),
    QStruct("distant_light_0_direction", INCLUDE=ijk_float),
    QStruct("distant_light_1_color", INCLUDE=rgb_float),
    QStruct("distant_light_1_direction", INCLUDE=ijk_float),
    Pad(12),

    QStruct("reflection_tint", INCLUDE=argb_float),
    QStruct("shadow_vector", INCLUDE=ijk_float),
    QStruct("shadow_color", INCLUDE=rgb_float),
    QStruct("plane", INCLUDE=plane),
    SInt16("breakable_surface", EDITABLE=False),
    Pad(6),

    SInt32("vertices_count", EDITABLE=False),
    SInt32("vertices_offset", EDITABLE=False, VISIBLE=False),

    FlUInt32("unknown_meta_offset0", VISIBLE=False),
    FlUInt32("vertices_meta_offset",
        TOOLTIP=("In xbox maps this is a bspmagic relative pointer that\n"
                 "points to a reflexive. The reflexive contains only a\n"
                 "bspmagic relative pointer to the vertices."),
        VISIBLE=False
        ),
    FlUEnum16("vertex_type",  # name is a guess
        ("unknown", 0),
        ("uncompressed", 2),
        ("compressed",   3),
        VISIBLE=False,
        ),
    Pad(2),
    SInt32("lightmap_vertices_count", EDITABLE=False),
    SInt32("lightmap_vertices_offset", EDITABLE=False, VISIBLE=False),

    FlUInt32("unknown_meta_offset1", VISIBLE=False),
    FlUInt32("lightmap_vertices_meta_offset",
        TOOLTIP=("In xbox maps this is a bspmagic relative pointer that\n"
                 "points to a reflexive. The reflexive contains only a\n"
                 "bspmagic relative pointer to the lightmap vertices."),
        VISIBLE=False
        ),

    rawdata_ref("uncompressed_vertices", max_size=4864000),
    rawdata_ref("compressed_vertices", max_size=2560000),
    SIZE=256
    )

lightmap = Struct("lightmap",
    SInt16("bitmap_index"),
    Pad(18),
    reflexive("materials", material, 2048,
        DYN_NAME_PATH='.shader.filepath'),
    SIZE=32
    )

lens_flare = Struct("lens_flare",
    dependency("shader", 'lens'),
    SIZE=16
    )

lens_flare_marker = Struct("lens_flare_marker",
    QStruct("position", INCLUDE=xyz_float),
    QStruct("direction",
        SInt8('i'), SInt8('j'), SInt8('k'), ORIENT='h'
        ),
    # While guerilla treats this like a signed int, there is no way that it
    # is gonna be able to reference one of the 256 lens flares if its signed
    UInt8('lens_flare_index'),
    SIZE=16
    )

surface_index = QStruct("surface_index",
    SInt32("surface_index"),
    SIZE=4
    )

mirror = Struct("mirror",
    QStruct("plane", INCLUDE=plane),
    # might be padding, might not be
    BytesRaw("unknown", VISIBLE=False, SIZE=20),
    #Pad(20),
    dependency("shader", valid_shaders),
    reflexive("vertices", vertex, 512),
    SIZE=64
    )

portal = QStruct("portal",
    SInt16("portal"),
    SIZE=2
    )

subcluster = Struct("subcluster",
    QStruct('world_bounds_x', INCLUDE=from_to),
    QStruct('world_bounds_y', INCLUDE=from_to),
    QStruct('world_bounds_z', INCLUDE=from_to),
    reflexive("surface_indices", surface_index, 128),
    SIZE=36, COMMENT="""
Subclusters define areas to render in square chunks. Surfaces indices are
the renderable surfaces(not collision) to render in this subcluster. This is
how Halo's renderer knows what surfaces to render in each cluster."""
    )

cluster = Struct("cluster",
    SInt16('sky'),
    SInt16('fog', TOOLTIP=cluster_fog_tooltip),
    dyn_senum16('background_sound',
        DYN_NAME_PATH="tagdata.background_sounds_palette.STEPTREE[DYN_I].name"),
    dyn_senum16('sound_environment',
        DYN_NAME_PATH="tagdata.sound_environments_palette." +
        "STEPTREE[DYN_I].name"),
    dyn_senum16('weather',
        DYN_NAME_PATH="tagdata.weather_palettes.STEPTREE[DYN_I].name"),

    UInt16("transition_structure_bsp", VISIBLE=False),
    UInt16("first_decal_index", VISIBLE=False),
    UInt16("decal_count", VISIBLE=False),

    # almost certain this is padding, though a value in the third
    # and fourth bytes is non-zero in meta, but not in a tag, so idk.
    Pad(24),

    reflexive("predicted_resources", predicted_resource, 1024, VISIBLE=False),
    reflexive("subclusters", subcluster, 4096),
    SInt16("first_lens_flare_marker_index"),
    SInt16("lens_flare_marker_count"),
    reflexive("surface_indices", surface_index, 32768),
    reflexive("mirrors", mirror, 16, DYN_NAME_PATH=".shader.filepath"),
    reflexive("portals", portal, 128),
    SIZE=104
    )

cluster_portal = Struct("cluster_portal",
    SInt16("front_cluster"),
    SInt16("back_cluster"),
    SInt32("plane_index"),
    QStruct("centroid", INCLUDE=xyz_float),
    Float("bounding_radius"),
    Bool32("flags",
        "ai_cant_hear_through_this",
        ),

    # might be padding, might not be
    BytesRaw("unknown", VISIBLE=False, SIZE=24),
    #Pad(24),
    reflexive("vertices", vertex, 128),
    SIZE=64
    )

breakable_surface = Struct("breakable_surface",
    QStruct("centroid", INCLUDE=xyz_float),
    Float("radius"),
    SInt32("collision_surface_index"),
    Pad(28),
    SIZE=48
    )

fog_plane = Struct("fog_plane",
    SInt16("front_region"),
    FlSEnum16("material_type",
        *(tuple((materials_list[i], i) for i in
           range(len(materials_list))) + (("NONE", -1), )),
        VISIBLE=False),  # calculated when compiled into map
    QStruct("plane", INCLUDE=plane),
    reflexive("vertices", vertex, 4096),
    SIZE=32
    )

fog_region = Struct("fog_region",
    Pad(36),
    dyn_senum16("fog_palette",
        DYN_NAME_PATH="tagdata.fog_palettes.STEPTREE[DYN_I].name"),
    dyn_senum16("weather_palette",
        DYN_NAME_PATH="tagdata.weather_palettes.STEPTREE[DYN_I].name"),
    SIZE=40
    )

fog_palette = Struct("fog_palette",
    ascii_str32("name"),
    dependency("fog", "fog "),
    Pad(4),
    ascii_str32("fog_scale_function"),
    SIZE=136
    )

weather_palette = Struct("weather_palette",
    ascii_str32("name"),
    dependency("particle_system", "rain"),
    Pad(4),
    ascii_str32("particle_system_scale_function"),

    Pad(44),
    dependency("wind", "wind"),
    QStruct("wind_direction", INCLUDE=ijk_float),
    Float("wind_magnitude"),
    Pad(4),
    ascii_str32("wind_scale_function"),
    SIZE=240
    )

weather_polyhedra = Struct("weather_polyhedra",
    QStruct("bounding_sphere_center", INCLUDE=xyz_float),
    Float("bounding_sphere_radius"),
    Pad(4),
    reflexive("planes", plane, 16),
    SIZE=32
    )

pathfinding_surface = QStruct("pathfinding_surface",
    # this is actually a 3bit width index, 3bit height
    # index, and 2 flags(is_walkable and is_breakable/is_broken)
    # stored in a single byte(in that order)
    UInt8("data"),
    SIZE=1
    )

pathfinding_edge = QStruct("pathfinding_edge", UInt8("midpoint"), SIZE=1)

background_sound_palette = Struct("background_sound_palette",
    ascii_str32("name"),
    dependency("background_sound", "lsnd"),
    Pad(4),
    ascii_str32("scale_function"),
    SIZE=116
    )

sound_environment_palette = Struct("sound_environment_palette",
    ascii_str32("name"),
    dependency("sound_environment", "snde"),
    SIZE=80
    )

marker = Struct("marker",
    ascii_str32("name"),
    QStruct("rotation", INCLUDE=ijkw_float),
    QStruct("position", INCLUDE=xyz_float),
    SIZE=60
    )


detail_object_cell = QStruct("detail_object_cell",
    SInt16("unknown1"), SInt16("unknown2"),
    SInt16("unknown3"), SInt16("unknown4"),
    SInt32("unknown5"), SInt32("unknown6"), SInt32("unknown7"),
    SIZE=32
    )

detail_object_instance = QStruct("detail_object_instance",
    SInt8("unknown1"), SInt8("unknown2"),
    SInt8("unknown3"), SInt8("unknown4"), SInt16("unknown5"),
    SIZE=6
    )

detail_object_count = QStruct("detail_object_count",
    SInt16("unknown"),
    SIZE=2
    )

detail_object_z_reference_vector = QStruct("detail_object_z_reference_vector",
    Float("unknown1"), Float("unknown2"),
    Float("unknown3"), Float("unknown4"),
    SIZE=16
    )

detail_object = Struct("detail_object",
    reflexive("cells", detail_object_cell, 262144),
    reflexive("instances", detail_object_instance, 2097152),
    reflexive("counts", detail_object_count, 8388608),
    reflexive("z_reference_vectors", detail_object_z_reference_vector, 262144),
    Bool8("flags",
        "enabled",  # required to be set on map compile.
        #             set to   "parent.instances.size != 0"
        VISIBLE=False
        ),
    SIZE=64
    )

runtime_decal = BytesRaw("unknown", SIZE=16)


face_vertex = QStruct("vertex", Float("x"), Float("y"), SIZE=8)
portal_index = Struct("portal_index", SInt32("portal_index"), SIZE=4)

face = Struct("face",
    SInt32("node_index"),
    reflexive("vertices", face_vertex, 64),
    SIZE=16
    )

leaf_map_leaf = Struct("leaf_map_leaf",
    reflexive("faces", face, 256),
    reflexive("portal_indices", portal_index, 256),
    SIZE=24
    )

leaf_map_portal = Struct("leaf_map_portal",
    SInt32("plane_index"),
    SInt32("back_leaf_index"),
    SInt32("front_leaf_index"),
    reflexive("vertices", face_vertex, 64),
    SIZE=24
    )

raw_cluster_data = QStruct("raw_cluster_data",
    UInt16("unknown0"),
    UInt16("unknown1"),
    UInt16("unknown2"),
    UInt16("unknown3"),
    SIZE=8
    )

sbsp_body = Struct("tagdata",
    dependency("lightmap_bitmaps", 'bitm'),
    float_wu("vehicle_floor"),  # world units
    float_wu("vehicle_ceiling"),  # world units

    Pad(20),
    QStruct("default_ambient_color", INCLUDE=rgb_float),
    Pad(4),
    QStruct("default_distant_light_0_color", INCLUDE=rgb_float),
    QStruct("default_distant_light_0_direction", INCLUDE=ijk_float),
    QStruct("default_distant_light_1_color", INCLUDE=rgb_float),
    QStruct("default_distant_light_1_direction", INCLUDE=ijk_float),

    Pad(12),
    QStruct("default_reflection_tint", INCLUDE=argb_float),
    QStruct("default_shadow_vector", INCLUDE=ijk_float),
    QStruct("default_shadow_color", INCLUDE=rgb_float),

    Pad(4),
    reflexive("collision_materials", collision_material, 512,
        DYN_NAME_PATH='.shader.filepath'),
    reflexive("collision_bsp", collision_bsp, 1),
    reflexive("nodes", node, 131072, VISIBLE=False),
    QStruct("world_bounds_x", INCLUDE=from_to),
    QStruct("world_bounds_y", INCLUDE=from_to),
    QStruct("world_bounds_z", INCLUDE=from_to),
    reflexive("leaves", leaf, 65535),
    reflexive("leaf_surfaces", leaf_surface, 262144),
    reflexive("surfaces", surface, 131072),
    reflexive("lightmaps", lightmap, 128),

    Pad(12),
    reflexive("lens_flares", lens_flare, 256,
        DYN_NAME_PATH='.shader.filepath'),
    reflexive("lens_flare_markers", lens_flare_marker, 65535),
    reflexive("clusters", cluster, 8192),

    # this is an array of 8 byte structs for each cluster
    rawdata_ref("cluster_data", max_size=65536),
    reflexive("cluster_portals", cluster_portal, 512),

    Pad(12),
    reflexive("breakable_surfaces", breakable_surface, 256),
    reflexive("fog_planes", fog_plane, 32),
    reflexive("fog_regions", fog_region, 32),
    reflexive("fog_palettes", fog_palette, 32,
        DYN_NAME_PATH='.name'),

    Pad(24),
    reflexive("weather_palettes", weather_palette, 32,
        DYN_NAME_PATH='.name'),
    reflexive("weather_polyhedras", weather_polyhedra, 32),

    Pad(24),
    reflexive("pathfinding_surfaces", pathfinding_surface, 131072, VISIBLE=False),
    reflexive("pathfinding_edges", pathfinding_edge, 262144, VISIBLE=False),
    reflexive("background_sounds_palette", background_sound_palette, 64,
        DYN_NAME_PATH='.name'),
    reflexive("sound_environments_palette", sound_environment_palette, 64,
        DYN_NAME_PATH='.name'),
    rawdata_ref("sound_pas_data", max_size=131072),

    UInt32("unknown", VISIBLE=False),
    Pad(20),
    reflexive("markers", marker, 1024, DYN_NAME_PATH='.name'),
    reflexive("detail_objects", detail_object, 1, VISIBLE=False),

    # the runtime decals reflexive is populated ONLY by the
    # engine while it is running(I'm making an educated guess)
    reflexive("runtime_decals", runtime_decal, 6144, VISIBLE=False),

    Pad(12),
    reflexive("leaf_map_leaves", leaf_map_leaf, 65536, VISIBLE=False),
    reflexive("leaf_map_portals", leaf_map_portal, 524288, VISIBLE=False),
    SIZE=648,
    )

fast_sbsp_body = dict(sbsp_body)
fast_sbsp_body[16] = reflexive("collision_bsp", fast_collision_bsp, 1)
fast_sbsp_body[17] = raw_reflexive("nodes", node, 131072)
fast_sbsp_body[21] = raw_reflexive("leaves", leaf, 65535)
fast_sbsp_body[22] = raw_reflexive("leaf_surfaces", leaf_surface, 262144)
fast_sbsp_body[23] = raw_reflexive("surface", surface, 131072)
fast_sbsp_body[27] = raw_reflexive("lens_flare_markers", lens_flare_marker, 65535)
fast_sbsp_body[32] = raw_reflexive("breakable_surfaces", breakable_surface, 256)
fast_sbsp_body[40] = raw_reflexive("pathfinding_surfaces", pathfinding_surface, 131072)
fast_sbsp_body[41] = raw_reflexive("pathfinding_edges", pathfinding_edge, 262144)
fast_sbsp_body[47] = raw_reflexive("markers", marker, 1024, DYN_NAME_PATH='.name')


sbsp_meta_header_def = BlockDef("sbsp_meta_header",
    # to convert these pointers to offsets, do:  pointer - bsp_magic
    UInt32("meta_pointer"),
    UInt32("uncompressed_lightmap_materials_count"),
    UInt32("uncompressed_lightmap_materials_pointer"),  # name is a guess
    UInt32("compressed_lightmap_materials_count"),
    UInt32("compressed_lightmap_materials_pointer"),  # name is a guess
    UInt32("sig", DEFAULT="sbsp"),
    SIZE=24, TYPE=QStruct
    )


def get():
    return sbsp_def

sbsp_def = TagDef("sbsp",
    blam_header("sbsp", 5),
    sbsp_body,

    ext=".scenario_structure_bsp", endian=">", tag_cls=SbspTag,
    )

fast_sbsp_def = TagDef("sbsp",
    blam_header("sbsp", 5),
    fast_sbsp_body,

    ext=".scenario_structure_bsp", endian=">", tag_cls=SbspTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Standardized
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

ugh__platform_codec_encoding = (
    "mono",
    "stereo",
    "surround",
    "_5_1_surround",
    )

ugh__platform_codec_sample_rate = (
    "_22khz",
    "_44khz",
    "_32khz",
    )


ugh__platform_codec = Struct("platform_codec",
    SEnum8("sample_rate", *ugh__platform_codec_sample_rate),
    SEnum8("encoding", *ugh__platform_codec_encoding),
    SInt8("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=3
    )


ugh__playback_parameter = Struct("playback_parameter",
    Bool32("field_disable_flags",
        "distance_a",
        "distance_b",
        "distance_c",
        "distance_d",
        ),
    Float("distance_a"),
    Float("distance_b"),
    Float("distance_c"),
    Float("distance_d"),
    Float("skip_fraction"),
    Float("maximum_bend_per_second"),
    Float("gain_base"),
    Float("gain_variance"),
    QStruct("random_pitch_bounds", INCLUDE=from_to_sint16),
    float_rad("inner_cone_angle"),
    float_rad("outer_cone_angle"),
    Float("outer_cone_gain"),
    Bool32("flags",
        "override_azimuth",
        "override_3d_gain",
        "override_speaker_gain",
        ),
    float_rad("azimuth"),
    Float("positional_gain"),
    Float("first_person_gain"),
    ENDIAN=">", SIZE=68
    )


ugh__scale = Struct("scale",
    QStruct("gain_modifier", INCLUDE=from_to),
    QStruct("pitch_modifier", INCLUDE=from_to_sint16),
    QStruct("skip_fraction_modifier", INCLUDE=from_to),
    ENDIAN=">", SIZE=20
    )


ugh__import_name = Struct("import_name",
    h3_string_id("name"),
    ENDIAN=">", SIZE=4
    )


ugh__pitch_range_parameter = Struct("pitch_range_parameter",
    SInt16("natural_pitch"),
    QStruct("bend_bounds", INCLUDE=from_to_sint16),
    QStruct("max_gain_pitch_bounds", INCLUDE=from_to_sint16),
    QStruct("unknown_bounds", INCLUDE=from_to_sint16, VISIBLE=False),
    ENDIAN=">", SIZE=14
    )


ugh__pitch_range = Struct("pitch_range",
    SInt16("import_name_index"),
    SInt16("pitch_range_parameter_index"),
    SInt16("encoded_permutation_data_index"),
    SInt16("encoded_runtime_permutation_flag_index"),
    SInt16("encoded_permutation_count"),
    UInt16("first_permuation_index"),
    ENDIAN=">", SIZE=12
    )


ugh__permutation = Struct("permutation",
    SInt16("import_name_index"),
    SInt16("encoded_skip_fraction"),
    UInt32("sample_size"),
    SInt32("first_permutation_chunk"),
    SInt16("permutation_chunk_count"),
    SInt8("gain"),
    SInt8("pitch_range_overall_permutation_index"),
    ENDIAN=">", SIZE=16
    )


ugh__custom_playback_filter = Struct("filter",
    SInt32("filter_type"),
    SInt32("filter_width"),
    QStruct("scale_bounds_0", INCLUDE=from_to),
    Float("random_base_0"),
    Float("random_variance_0"),
    QStruct("scale_bounds_1", INCLUDE=from_to),
    Float("random_base_1"),
    Float("random_variance_1"),
    QStruct("scale_bounds_2", INCLUDE=from_to),
    Float("random_base_2"),
    Float("random_variance_2"),
    QStruct("scale_bounds_3", INCLUDE=from_to),
    Float("random_base_3"),
    Float("random_variance_3"),
    ENDIAN=">", SIZE=72
    )


ugh__custom_playback = Struct("custom_playback",
    BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
    Bool32("flags",
        "use_3d_radio_hack",
        ),
    BytesRaw("unknown_1", SIZE=8, VISIBLE=False),
    h3_reflexive("filter", ugh__custom_playback_filter),
    BytesRaw("unknown_2", SIZE=36, VISIBLE=False),
    ENDIAN=">", SIZE=72
    )


ugh__language_unknown_a = Struct("unknown_a",
    SInt16("unknown"),
    ENDIAN=">", SIZE=2
    )


ugh__language_unknown_b = Struct("unknown_b",
    SInt16("unknown_a_start_index"),
    SInt16("unknown_a_count"),
    ENDIAN=">", SIZE=4
    )


ugh__language = Struct("language",
    SEnum32("language", *ugh__language_language),
    h3_reflexive("unknown_a", ugh__language_unknown_a),
    h3_reflexive("unknown_b", ugh__language_unknown_b),
    ENDIAN=">", SIZE=28
    )


ugh__runtime_permutation_flag = Struct("runtime_permutation_flag",
    SInt8("unknown"),
    ENDIAN=">", SIZE=1
    )


ugh__permutation_chunk = Struct("permutation_chunk",
    UInt32("file_offset"),
    Bool8("flags",
        ("has_unknown_a_value", 1 << 5),
        ),
    UInt8("chunk_size_leftmost_byte"),
    UInt16("chunk_size"),
    SInt32("runtime_index"),
    SInt32("unknown_a"),
    SInt32("unknown_size"),
    ENDIAN=">", SIZE=20
    )


ugh__promotion_rule = Struct("rule",
    SInt16("pitch_range_index"),
    SInt16("maximum_playing_count"),
    Float("suppression_time"),
    SInt32("unknown_0", VISIBLE=False),
    SInt32("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


ugh__promotion_runtime_timer = Struct("runtime_timer",
    SInt32("unknown"),
    ENDIAN=">", SIZE=4
    )


ugh__promotion = Struct("promotion",
    h3_reflexive("rules", ugh__promotion_rule),
    h3_reflexive("runtime_timers", ugh__promotion_runtime_timer),
    SInt32("unknown_0", VISIBLE=False),
    BytesRaw("unknown_1", SIZE=8, VISIBLE=False),
    ENDIAN=">", SIZE=36
    )


ugh__extra_info_encoded_permutation_section_sound_dialogue_info = Struct("sound_dialogue_info",
    UInt32("mouth_data_offset"),
    UInt32("mouth_data_length"),
    UInt32("lipsync_data_offset"),
    UInt32("lipsynce_data_length"),
    ENDIAN=">", SIZE=16
    )


ugh__extra_info_encoded_permutation_section_unknown_unknown_unknown_1 = Struct("unknown_1",
    Pad(8),
    ENDIAN=">", SIZE=8
    )


ugh__extra_info_encoded_permutation_section_unknown_unknown_unknown_2 = Struct("unknown_2",
    Array("unknown_array", SUB_STRUCT=SInt16("unknown"), SIZE=4),
    ENDIAN=">", SIZE=8
    )


ugh__extra_info_encoded_permutation_section_unknown_unknown = Struct("unknown",
    Pad(16),
    h3_reflexive("unknown_1", ugh__extra_info_encoded_permutation_section_unknown_unknown_unknown_1),
    h3_reflexive("unknown_2", ugh__extra_info_encoded_permutation_section_unknown_unknown_unknown_2),
    ENDIAN=">", SIZE=40
    )


ugh__extra_info_encoded_permutation_section_unknown = Struct("unknown",
    h3_reflexive("unknown", ugh__extra_info_encoded_permutation_section_unknown_unknown),
    ENDIAN=">", SIZE=12
    )


ugh__extra_info_encoded_permutation_section = Struct("encoded_permutation_section",
    h3_rawdata_ref("encoded_data"),
    h3_reflexive("sound_dialogue_info", ugh__extra_info_encoded_permutation_section_sound_dialogue_info),
    h3_reflexive("unknown", ugh__extra_info_encoded_permutation_section_unknown),
    ENDIAN=">", SIZE=44
    )


ugh__extra_info = Struct("extra_info",
    h3_reflexive("encoded_permutation_sections", ugh__extra_info_encoded_permutation_section),
    ENDIAN=">", SIZE=12
    )


ugh__body = Struct("tagdata",
    h3_reflexive("platform_codec", ugh__platform_codec),
    h3_reflexive("playback_parameters", ugh__playback_parameter),
    h3_reflexive("scales", ugh__scale),
    h3_reflexive("import_names", ugh__import_name),
    h3_reflexive("pitch_range_parameters", ugh__pitch_range_parameter),
    h3_reflexive("pitch_ranges", ugh__pitch_range),
    h3_reflexive("permutations", ugh__permutation),
    h3_reflexive("custom_playbacks", ugh__custom_playback),
    h3_reflexive("languages", ugh__language),
    h3_reflexive("runtime_permutation_flags", ugh__runtime_permutation_flag),
    h3_rawdata_ref("unknown_0", VISIBLE=False),
    BytesRaw("unknown_1", SIZE=8, VISIBLE=False),
    h3_reflexive("permutation_chunks", ugh__permutation_chunk),
    h3_reflexive("promotions", ugh__promotion),
    h3_reflexive("extra_infos", ugh__extra_info),
    ENDIAN=">", SIZE=184
    )


def get():
    return ugh__def

ugh__def = TagDef("ugh!",
    h3_blam_header('ugh!'),
    ugh__body,

    ext=".%s" % h3_tag_class_fcc_to_ext["ugh!"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

wind_body = Struct("tagdata",
    from_to_wu("velocity"),
    yp_float_rad("variation_area"),
    Float("local_variation_weight"),
    Float("local_variation_rate"),
    Float("damping"),
    SIZE=64,
    )


def get():
    return wind_def

wind_def = TagDef("wind",
    blam_header('wind'),
    wind_body,

    ext=".wind", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = ( 'JmsTriangle', )


class JmsTriangle:
    __slots__ = (
        "region", "shader",
        "v0", "v1", "v2"
        )
    def __init__(self, region=0, shader=0,
                 v0=0, v1=0, v2=0):
        self.region = region
        self.shader = shader
        self.v0 = v0
        self.v1 = v1
        self.v2 = v2

    def __getitem__(self, index):
        if   index == 0: return self.v0
        elif index == 1: return self.v1
        elif index == 2: return self.v2
        raise IndexError("Triangle indices must be in range(0, 3)")

    def __repr__(self):
        return """JmsTriangle(
    region=%s, shader=%s,
    v0=%s, v1=%s, v2=%s
)""" % (self.region, self.shader,
        self.v0, self.v1, self.v2)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import math
import traceback

from reclaimer.sounds import util, constants

__all__ = ("compile_sound", "compile_pitch_range",)


def compile_pitch_range(pitch_range, blam_pitch_range,
                        channel_count, sample_rate,
                        ignore_size_limits=False,
                        force_sample_rate=False):
    '''
    Compiles the provided BlamSoundPitchRange into the
    provided snd! pitch range block.
    '''
    errors = []

    # make sure the imported sounds are the correct
    # encoding, sample rate, and have a valid compression.
    for name, blam_sound_perm in blam_pitch_range.permutations.items():
        blam_channel_count = constants.channel_counts.get(
            blam_sound_perm.encoding, "invalid")

        if not force_sample_rate and sample_rate != blam_sound_perm.sample_rate:
            errors.append('Cannot add %skHz sounds to %skHz tag.' %
                          (blam_sound_perm.sample_rate, sample_rate))

        if channel_count != blam_channel_count:
            errors.append('Cannot add %s channel sounds to %s channel tag.' %
                          (blam_channel_count, sample_rate))

        if blam_sound_perm.compression not in (constants.COMPRESSION_PCM_16_LE,
                                               constants.COMPRESSION_PCM_16_BE,
                                               constants.COMPRESSION_XBOX_ADPCM,
                                               constants.COMPRESSION_IMA_ADPCM,
                                               constants.COMPRESSION_OGG):
            errors.append('Unknown permutation compression "%s"' %
                          blam_sound_perm.compression)

        for blam_samples in blam_sound_perm.processed_samples:
            if len(blam_samples.sample_data) > constants.MAX_SAMPLE_CHUNK_SIZE:
                errors.append('Sample data too large to compile in "%s"' % name)

            if len(blam_samples.mouth_data) > constants.MAX_MOUTH_DATA:
                errors.append('Mouth data too large to compile in "%s"' % name)

    if errors:
        return errors

    snd__perms = pitch_range.permutations.STEPTREE
    snd__perm_names = set()
    # loop over the permutations blam_samples and string
    # them together into lists of permutation blocks
    for blam_perm_name in sorted(blam_pitch_range.permutations):
        blam_sound_perm = blam_pitch_range.permutations[blam_perm_name]
        name = get_unique_name_31char(
            blam_sound_perm.name.strip(), snd__perm_names)
        snd__perm_names.add(name)

        snd__perm_chain = []
        try:
            for blam_samples in blam_sound_perm.processed_samples:
                snd__perms.append()  # create the new perm block
                snd__perm, _ = snd__perms.pop(-1)
                snd__perm.name = name
                snd__perm.ogg_sample_count = (
                    channel_count * 2 * blam_samples.sample_count)
                snd__perm.samples.data = blam_samples.sample_data
                snd__perm.mouth_data.data = blam_samples.mouth_data

                if blam_samples.compression == constants.COMPRESSION_XBOX_ADPCM:
                    snd__perm.compression.set_to("xbox_adpcm")
                    snd__perm.ogg_sample_count = 0  # adpcm has this as 0 always
                elif blam_samples.compression == constants.COMPRESSION_IMA_ADPCM:
                    snd__perm.compression.set_to("ima_adpcm")
                    snd__perm.ogg_sample_count = 0  # adpcm has this as 0 always
                elif blam_samples.compression == constants.COMPRESSION_OGG:
                    snd__perm.compression.set_to("ogg")
                else:
                    snd__perm.compression.set_to("none")

                snd__perm_chain.append(snd__perm)

            errors.extend(
                merge_permutation_chain(
                    snd__perm_chain, pitch_range, ignore_size_limits)
                )
        except Exception:
            errors.append(traceback.format_exc())
            errors.append("Could not compile permutation %r" % blam_perm_name)

    return errors


def compile_sound(snd__tag, blam_sound_bank, ignore_size_limits=False,
                  update_mode=constants.SOUND_COMPILE_MODE_PRESERVE,
                  force_sample_rate=False):
    '''
    Compiles the given blam_sound_bank into the given (Halo 1 snd!) snd__tag.
    '''
    assert update_mode in (constants.SOUND_COMPILE_MODE_NEW,
                           constants.SOUND_COMPILE_MODE_PRESERVE,
                           constants.SOUND_COMPILE_MODE_ADDITIVE)
    errors = []
    tagdata = snd__tag.data.tagdata
    if update_mode == constants.SOUND_COMPILE_MODE_NEW:
        # clear the old contents of the tag
        tagdata.parse()

    forced_natural_pitch = 0
    if update_mode == constants.SOUND_COMPILE_MODE_NEW:
        # set some nice default values
        tagdata.random_pitch_bounds[:] = (1.0, 1.0)
        tagdata.inner_cone_angle = 2*math.pi
        tagdata.outer_cone_angle = 2*math.pi
        tagdata.outer_cone_gain = 1.0

        tagdata.modifiers_when_scale_is_zero[:] = (1.0, 0.0, 1.0)
        tagdata.modifiers_when_scale_is_one[:]  = (1.0, 1.0, 1.0)


    if update_mode != constants.SOUND_COMPILE_MODE_ADDITIVE:
        # update the flags, compression, encoding, and sample rate
        # of the tag to that of the samples being stored in it.
        tagdata.flags.fit_to_adpcm_blocksize = False
        tagdata.flags.split_long_sound_into_permutations = bool(
            blam_sound_bank.split_into_smaller_chunks)

        if blam_sound_bank.compression in (constants.COMPRESSION_PCM_16_LE,
                                           constants.COMPRESSION_PCM_16_BE):
            # intentionally set this to something other than "none"
            # as otherwise the sound won't play in sapien
            tagdata.compression.set_to("ogg")
            # tagdata.compression.set_to("none")
        elif blam_sound_bank.compression == constants.COMPRESSION_XBOX_ADPCM:
            tagdata.flags.fit_to_adpcm_blocksize = True
            tagdata.compression.set_to("xbox_adpcm")
        elif blam_sound_bank.compression == constants.COMPRESSION_IMA_ADPCM:
            tagdata.flags.fit_to_adpcm_blocksize = True
            tagdata.compression.set_to("ima_adpcm")
        elif blam_sound_bank.compression == constants.COMPRESSION_OGG:
            tagdata.compression.set_to("ogg")
        else:
            errors.append('Unknown compression "%s"' %
                          blam_sound_bank.compression)

        if blam_sound_bank.encoding == constants.ENCODING_MONO:
            tagdata.encoding.set_to("mono")
        elif blam_sound_bank.encoding == constants.ENCODING_STEREO:
            tagdata.encoding.set_to("stereo")
        else:
            errors.append('Unknown encoding "%s"' %
                          blam_sound_bank.encoding)

        if (blam_sound_bank.sample_rate == constants.SAMPLE_RATE_22K or
                force_sample_rate):
            if force_sample_rate:
                forced_natural_pitch = (constants.SAMPLE_RATE_22K /
                                        blam_sound_bank.sample_rate)

            tagdata.sample_rate.set_to("khz_22")
        elif blam_sound_bank.sample_rate == constants.SAMPLE_RATE_44K:
            tagdata.sample_rate.set_to("khz_44")
        else:
            errors.append('Unsupported sample rate "%s"' %
                          blam_sound_bank.sample_rate)

    elif force_sample_rate:
        errors.append('Cannot force sample rate in additive compile mode.')


    if errors:
        return errors

    # make sure the encoding and sample rate of the
    # BlamSoundBank match the settings in the tagdata.
    snd__sample_rate = constants.halo_1_sample_rates.get(
        tagdata.sample_rate.data, "invalid")
    snd__channel_count = constants.channel_counts.get(
        tagdata.encoding.data, "invalid")
    blam_channel_count = constants.channel_counts.get(
        blam_sound_bank.encoding, "invalid")

    if not force_sample_rate and snd__sample_rate != blam_sound_bank.sample_rate:
        errors.append('Cannot add %skHz sounds to %skHz tag.' %
                      (blam_sound_bank.sample_rate, snd__sample_rate))

    if snd__channel_count != blam_channel_count:
        errors.append('Cannot add %s channel sounds to %s channel tag.' %
                      (blam_channel_count, snd__channel_count))


    if errors:
        return errors

    snd__pitch_ranges = tagdata.pitch_ranges.STEPTREE
    blam_pitch_ranges = blam_sound_bank.pitch_ranges
    snd__pitch_ranges_by_name = {}
    prev_snd__pitch_ranges_by_name = {
        pr.name.lower().strip(): pr
        for pr in snd__pitch_ranges
        }

    encoding = tagdata.encoding.data
    sample_rate = constants.halo_1_sample_rates.get(
        tagdata.sample_rate.data, 0)

    # loop over the blam_sound_bank pitch ranges and compile them
    for blam_pr_name in sorted(blam_pitch_ranges):
        blam_pr = blam_pitch_ranges[blam_pr_name]
        name = get_unique_name_31char(
            blam_pr.name.strip(), snd__pitch_ranges_by_name)

        snd__pitch_ranges.append()  # create the new pitch range block
        snd__pitch_ranges_by_name[name], _ = snd__pitch_ranges.pop(-1)
        snd__pitch_ranges_by_name[name].name = blam_pr_name
        try:
            errors.extend(compile_pitch_range(
                snd__pitch_ranges_by_name[name], blam_pr,
                snd__channel_count, snd__sample_rate,
                ignore_size_limits, force_sample_rate))
        except Exception:
            errors.append(traceback.format_exc())
            errors.append("Could not compile pitch range '%s'" % blam_pr.name)
            snd__pitch_ranges_by_name.pop(name)


    if update_mode != constants.SOUND_COMPILE_MODE_NEW:
        # preserve any old tag values and/or permutations
        for snd__pr_name, snd__pr in snd__pitch_ranges_by_name.items():
            prev_snd__pr = prev_snd__pitch_ranges_by_name.get(snd__pr_name)
            if prev_snd__pr is None:
                continue

            snd__pr.natural_pitch = prev_snd__pr.natural_pitch
            snd__pr.bend_bounds = prev_snd__pr.bend_bounds

            prev_perms = prev_snd__pr.permutations.STEPTREE
            perms = snd__pr.permutations.STEPTREE

            prev_perm_count = max(0, min(
                len(prev_perms), prev_snd__pr.actual_permutation_count))
            perm_count = max(0, min(
                len(perms), snd__pr.actual_permutation_count))

            # sort permutations by name so we can copy values
            perm_indices_by_name = {
                perm.name.strip().lower(): i
                for i, perm in enumerate(perms[: perm_count])
                }

            indices_to_merge = set()
            # copy values from old permutations into new ones
            for i, prev_perm in enumerate(prev_perms[: prev_perm_count]):
                perm_name = prev_perm.name.strip().lower()

                if perm_name not in perm_indices_by_name:
                    # no matching new perm. queue this one to be merged
                    indices_to_merge.add(i)
                    continue

                perm_pieces = get_permutation_chain(
                    perm_indices_by_name[perm_name], snd__pr)
                if not perm_pieces:
                    continue

                for perm_piece in perm_pieces:
                    perm_piece.name = prev_perm.name
                    perm_piece.gain = prev_perm.gain
                    perm_piece.skip_fraction = 0.0

                perm_pieces[0].skip_fraction = prev_perm.skip_fraction


            if update_mode != constants.SOUND_COMPILE_MODE_ADDITIVE:
                continue

            # merge old permutations into new pitch range
            for i in indices_to_merge:
                errors.extend(
                    merge_permutation_chain(
                        get_permutation_chain(i, prev_snd__pr),
                        snd__pr, ignore_size_limits)
                    )


    if update_mode == constants.SOUND_COMPILE_MODE_ADDITIVE:
        # add back in any pitch ranges missing from the old tag
        snd__pitch_ranges_by_name.update(
            {name: pr for name, pr in
             prev_snd__pitch_ranges_by_name.items()
             if name not in snd__pitch_ranges_by_name}
            )

    # clear the pitch ranges and update with the new ones
    del snd__pitch_ranges[:]
    for name in sorted(snd__pitch_ranges_by_name):
        pitch_range = snd__pitch_ranges_by_name[name]
        if forced_natural_pitch:
            pitch_range.natural_pitch = forced_natural_pitch

        if not pitch_range.permutations.STEPTREE:
            continue

        if len(snd__pitch_ranges) < snd__pitch_ranges.MAX or ignore_size_limits:
            snd__pitch_ranges.append(pitch_range)
        else:
            errors.append("Too many pitch ranges. Cannot add '%s'" % name)

    return errors


def get_unique_name_31char(name, names):
    name = name[: 31]
    if name not in names:
        return name

    i = 1
    end_str = ""
    while name[: 31 - len(end_str)] + end_str in names:
        end_str = "~" + str(i)
        i += 1

    return name[: 31 - len(end_str)] + end_str


def get_permutation_chain(start_index, pitch_range):

    seen = set()
    chain = []
    perms = pitch_range.permutations.STEPTREE
    i = start_index
    while i in range(len(perms)) and i not in seen:
        seen.add(i)
        chain.append(perms[i])
        i = perms[i].next_permutation_index

    return chain


def merge_permutation_chain(perm_chain, pitch_range, ignore_size_limits=False):
    '''
    What?
    '''
    errors = []
    if not perm_chain:
        return errors

    perms = pitch_range.permutations.STEPTREE
    perm_count = max(0, min(len(perms), pitch_range.actual_permutation_count))
    if ignore_size_limits:
        # signed integer max is because the next_permutation_index value is a SInt16
        if perm_count + len(perm_chain) >= 0x8000:
            errors.append("Too many permutations. Cannot add '%s'" %
                          perm_chain[0].name)
    elif len(perms) + len(perm_chain) > perms.MAX:
        errors.append("Too many permutation pieces. Cannot add '%s'" %
                      perm_chain[0].name)

    if errors:
        return errors

    for perm in perms:
        # shift all the next perm indices to account for the added perm
        if perm.next_permutation_index >= perm_count:
            perm.next_permutation_index += 1

    # insert the first piece of the permutation chain into the perms
    pitch_range.actual_permutation_count += 1
    perms.insert(perm_count, perm_chain[0])
    perm_chain[0].next_permutation_index = len(perms)

    # append and set the next index for each piece in the chain
    for perm in perm_chain[1: ]:
        perms.append(perm)
        perm.next_permutation_index = len(perms)

    # cut off the chain at last piece
    perm_chain[-1].next_permutation_index = -1

    return errors



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from ...hek.defs.objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

infection_form = Struct("infection_form",
    dependency_os("infection_form", valid_devices_items_objects_units),
    SIZE=16
    )

infectable_unit = Struct("infectable_unit",
    Pad(2),
    dyn_senum16("infection_form",
        DYN_NAME_PATH=".....infection_forms.STEPTREE" +
        "[DYN_I].infection_form.filepath"),
    dependency_os("infectable_unit", valid_units),
    Float("health_threshold"),
    dependency_os("infected_unit", valid_units),
    dependency_os("infected_unit_actor_variant", "actv"),
    dependency_os("transition_effect", "effe"),
    dependency_os("attachment_object", valid_devices_items_objects_units),
    ascii_str32("attachment_marker"),
    SInt16("attachment_marker_count"),
    SIZE=144
    )

unit_infection = Struct("unit_infection",
    Pad(4),
    reflexive("infection_forms", infection_form, 8,
        DYN_NAME_PATH='.infection_form.filepath'),
    reflexive("infectable_units", infectable_unit, 64,
        DYN_NAME_PATH='.infectable_unit.filepath'),
    SIZE=52
    )

boarding_seat = Struct("boarding_seat",
    Bool16("flags",
        "boarding_ejects_target_seat",
        "boarding_enters_target_seat",
        "controls_open_and_close",
        ),
    Pad(6),
    ascii_str32("seat_label"),
    ascii_str32("target_seat_label"),
    dependency_os("boarding_damage", "jpt!"),
    SIZE=120
    )

unit_external_upgrade = Struct("unit_external_upgrade",
    Pad(4),
    dependency_os("unit", valid_units),
    reflexive("boarding_seats", boarding_seat, 32,
        DYN_NAME_PATH='.seat_label'),
    SIZE=68
    )

gelc_body = Struct("tagdata",
    SInt16("version"),
    Bool16("flags",
        "allow_unit_infections_during_cinematics",
        ),
    reflexive("unit_infections", unit_infection, 1),
    reflexive("unit_external_upgrades", unit_external_upgrade, 64,
        DYN_NAME_PATH='.unit.filepath'),
    SIZE=172
    )

def get():
    return gelc_def

gelc_def = TagDef("gelc",
    blam_header_os('gelc'),
    gelc_body,

    ext=".project_yellow_globals_cv", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os
from array import array
from math import sqrt


class P8Palette():
    # Maps [(red << 8) + green]  to its palette index,
    # where the red/green pairs are restricted to
    # being those found in the 256 color palette.
    # and are in the range of [0, 255]
    # Contains EXACTLY 2^8 entries
    palette_rg = ()

    # Maps [(red << 8) + green]  to its palette index,
    # where red and green are independent of one another
    # and are in the range of [0, 255]
    # Contains EXACTLY 2^16 entries
    palette_map = ()

    palette_directory = os.path.dirname(__file__)
    palette_filename = ""

    palette_map_filename = ""

    def __init__(self, palette_filename=""):
        if palette_filename:
            self.palette_filename = palette_filename

        self.palette_map_filename = self.palette_filename + "_diff_map"

        with open(os.path.join(self.palette_directory,
                               self.palette_filename), 'rb') as f:
            palette_data = f.read(1024)

        # now build the array to convert p8 back to 32 bit
        self.p8_palette_32bit = unpacked = array("B", palette_data)
        self.p8_palette_32bit_packed = packed = array("L")
        self.palette_rg = array("H", [0] * 256)
        for i in range(256):
            a, r, g, b = unpacked[i * 4: i * 4 + 4]
            self.palette_rg[i] = (r << 8) + g
            packed.append((a<<24) + (r<<16) + (g<<8) + b)

        self.load_palette_map(False)

    @property
    def palette_map_loaded(self):
        return self.palette_map and len(self.palette_map) == 65536

    def write_palette_image(self):
        width = height = 256
        with open(os.path.join(self.palette_directory,
                               self.palette_filename + ".bmp"), 'wb+') as f:
            base = 138
            size = base + (width * height * 4)
            f.write(b"BM" + size.to_bytes(4, 'little'))
            f.write(b"\x00\x00\x00\x00\x8A\x00\x00\x00\x7C\x00\x00\x00")
            f.write(width.to_bytes(4, 'little') + height.to_bytes(4, 'little'))
            f.write(b"\x01\x00\x20\x00")
            f.write(b'\x00' * (size - f.tell()))
            for r in range(256):
                for g in range(256):
                    f.seek((g * width + r) * 4 + base)
                    color = (r << 8) + g
                    if color in self.palette_rg:
                        continue
                    i = self.palette_map[color] * 4
                    f.write(self.p8_palette_32bit[i: i + 4][::-1])

    def load_palette_map(self, generate_on_fail=True):
        try:
            with open(os.path.join(self.palette_directory,
                                   self.palette_map_filename), 'rb') as f:
                data = f.read(65536)
                if len(data) != 65536:
                    raise ValueError("Cached smallest diff map is invalid.")
                self.palette_map = list(data)
        except Exception:
            if generate_on_fail:
                self.generate_palette_map()
                try:
                    self.cache_palette_map()
                except Exception:
                    pass

    def cache_palette_map(self):
        with open(os.path.join(self.palette_directory,
                               self.palette_map_filename), 'wb+') as f:
            f.write(array("B", self.palette_map))

    def generate_palette_map(self):
        diff_map = [65536] * 65536
        palette_map = self.palette_map = [0] * 65536
        range_256 = tuple(range(256))

        colors = list(((color >> 8) & 0xFFff)
                      for color in self.p8_palette_32bit_packed)
        for i in range(len(colors)):
            if colors[i] == 0:
                continue

            pal_r = colors[i] >> 8
            pal_g = colors[i] & 0xFF
            for r in range_256:
                r_diff = (pal_r - r) ** 2
                r = r << 8
                for g in range_256:
                    diff = r_diff + (pal_g - g) ** 2
                    true_rg = r | g
                    if diff < diff_map[true_rg]:
                        diff_map[true_rg] = diff
                        palette_map[true_rg] = i

    def argb_array_to_p8_array_best_fit_alpha(self, unpacked_pix):
        if not self.palette_map_loaded:
            self.load_palette_map()

        color_map = self.palette_map
        indexing = array("B", bytearray(len(unpacked_pix)//4))
        for i in range(0, len(unpacked_pix), 4):
            if unpacked_pix[i]:
                red = unpacked_pix[i+1]
                green = unpacked_pix[i+2]
                indexing[i>>2] = color_map[(red<<8)+green]
            else:
                indexing[i>>2] = 255

        return (self.p8_palette_32bit, indexing)

    def argb_array_to_p8_array_best_fit(self, unpacked_pix):
        if not self.palette_map_loaded:
            self.load_palette_map()

        color_map = self.palette_map
        indexing = array("B", bytearray(len(unpacked_pix)//4))
        for i in range(0, len(unpacked_pix), 4):
            red = unpacked_pix[i+1]
            green = unpacked_pix[i+2]
            indexing[i>>2] = color_map[(red<<8)+green]

        return (self.p8_palette_32bit, indexing)


HALO_P8_PALETTE = P8Palette("p8_palette_halo")
STUBBS_P8_PALETTE = P8Palette("p8_palette_stubbs")



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Woo
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


sgp__body = Struct("tagdata",
    h3_dependency("underwater_environment"),
    h3_dependency("underwater_loop"),
    BytesRaw("unknown", SIZE=8, VISIBLE=False),
    h3_dependency("enter_underater"),
    h3_dependency("exit_underwater"),
    ENDIAN=">", SIZE=72
    )


def get():
    return sgp__def

sgp__def = TagDef("sgp!",
    h3_blam_header('sgp!'),
    sgp__body,

    ext=".%s" % h3_tag_class_fcc_to_ext["sgp!"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...os_v3_hek.defs.mach import *

#import and use the open saucified obje attrs
from .obje import *

# replace the object_type enum one that uses
# the correct default value for this object
obje_attrs = dict(obje_attrs)
obje_attrs[0] = dict(obje_attrs[0], DEFAULT=7)

mach_body = dict(mach_body)
mach_body[0] = obje_attrs

def get():
    return mach_def

mach_def = TagDef("mach",
    blam_header('mach'),
    mach_body,

    ext=".device_machine", endian=">", tag_cls=MachTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: DeadCanadian
# 	naming stuff
# revision: 3		author: OrangeMohawk
# 	Responses/Indexes
# revision: 4		author: Lord Zedd
# 	Updates and corrections
# revision: 5		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

adlg_vocalization_perception_type = (
    "none",
    "speaker",
    "listener",
    )


adlg_vocalization_response = Struct("response",
    h3_string_id("vocalization_name"),
    Bool16("flags", *unknown_flags_16),
    SInt16("vocalization_index"),
    SInt16("response_type"),
    SInt16("import_dialogue_index"),
    ENDIAN=">", SIZE=12
    )


adlg_vocalization = Struct("vocalization",
    h3_string_id("vocalization"),
    SInt16("parent_index"),
    SInt16("priority"),
    Bool32("flags", *unknown_flags_32),
    SInt16("glance_behavior"),
    SInt16("glance_recipient"),
    SEnum16("perception_type", *adlg_vocalization_perception_type),
    SInt16("max_combat_status"),
    SInt16("animation_impulse"),
    SInt16("overlap_priority"),
    Float("sound_repetition_delay"),
    Float("allowable_queue_delay"),
    Float("pre_vocalization_delay"),
    Float("notification_delay"),
    Float("post_vocalization_delay"),
    Float("repeat_delay"),
    Float("weight"),
    Float("speaker_freeze_time"),
    Float("listener_freeze_time"),
    SInt16("speaker_emotion"),
    SInt16("listener_emotion"),
    Float("skip_fraction1"),
    Float("skip_fraction2"),
    Float("skip_fraction3"),
    h3_string_id("sample_line"),
    h3_reflexive("responses", adlg_vocalization_response),
    ENDIAN=">", SIZE=92
    )


adlg_pattern = Struct("pattern",
    SInt16("dialogue_type"),
    SInt16("vocalizations_index"),
    h3_string_id("vocalization_name"),
    SInt16("speaker_type"),
    Bool16("flags", *unknown_flags_16),
    SInt16("hostility"),
    Bool16("unknown_0", *unknown_flags_16),
    SInt16("unknown_1", VISIBLE=False),
    SInt16("cause_type"),
    h3_string_id("cause_ai_type_name"),
    BytesRaw("unknown_2", SIZE=4, VISIBLE=False),
    SInt16("unknown_3", VISIBLE=False),
    SInt16("unknown_4", VISIBLE=False),
    SInt16("attitude"),
    SInt16("unknown_5", VISIBLE=False),
    Bool32("conditions", *unknown_flags_32),
    SInt16("spacial_relationship"),
    SInt16("damage_type"),
    SInt16("unknown_6", VISIBLE=False),
    SInt16("subject_type"),
    h3_string_id("subject_ai_type_name"),
    ENDIAN=">", SIZE=52
    )


adlg_dialog_data = Struct("dialog_data",
    SInt16("start_index"),
    SInt16("length"),
    ENDIAN=">", SIZE=4
    )


adlg_involuntary_data = Struct("involuntary_data",
    SInt16("involuntary_vocalization_index"),
    SInt16("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=4
    )


adlg_body = Struct("tagdata",
    BytesRaw("unknown_0", SIZE=16, VISIBLE=False),
    h3_reflexive("vocalizations", adlg_vocalization),
    h3_reflexive("patterns", adlg_pattern),
    BytesRaw("unknown_1", SIZE=12, VISIBLE=False),
    h3_reflexive("dialog_data", adlg_dialog_data),
    h3_reflexive("involuntary_data", adlg_involuntary_data),
    ENDIAN=">", SIZE=76
    )


def get():
    return adlg_def

adlg_def = TagDef("adlg",
    h3_blam_header('adlg'),
    adlg_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["adlg"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

velocity_properties = Struct("velocity_properties",
    SEnum16("attachment", *function_outputs),
    Pad(2),
    Float("velocity", UNIT_SCALE=per_sec_unit_scale),
    Float("low_multiplier"),
    Float("high_multiplier")
    )

glw__body = Struct("tagdata",
    ascii_str32("attachment_marker"),
    SInt16("number_of_particles"),
    SEnum16("boundary_effect",
        "bounce",
        "wrap",
        ),
    SEnum16("normal_particle_distribution",
        "random",
        "uniform",
        ),
    SEnum16("trailing_particle_distribution",
        "vertically",
        "normal",
        "randomly"
        ),
    Bool32("glow_flags" ,
        "modify_particle_color_in_range",
        "particles_move_backwards",
        "particles_move_in_both_directions",
        "trailing_particles_fade_over_time",
        "trailing_particles_shrink_over_time",
        "trailing_particles_slow_over_time",
        ),

    Pad(36),
    Struct("particle_rotational_velocity", INCLUDE=velocity_properties),
    Struct("effect_rotational_velocity", INCLUDE=velocity_properties),
    Struct("effect_translational_velocity", INCLUDE=velocity_properties),
    Struct("particle_distance_to_object",
        SEnum16("attachment", *function_outputs),
        Pad(2),
        Float("min_distance"),
        Float("max_distance"),
        Float("low_multiplier"),
        Float("high_multiplier")
        ),

    Pad(8),
    Struct("particle_size",
        SEnum16("attachment", *function_outputs),
        Pad(2),
        from_to_wu("size_bounds"),  # world units
        QStruct("size_attachment_multiplier", INCLUDE=from_to),
        ),

    Struct("particle_color",
        SEnum16("attachment", *function_outputs),
        Pad(2),
        QStruct("lower_bound", INCLUDE=argb_float),
        QStruct("upper_bound", INCLUDE=argb_float),
        QStruct("lower_scale", INCLUDE=argb_float),
        QStruct("upper_scale", INCLUDE=argb_float)
        ),

    Float("color_rate_of_change"),
    Float("fading_percentage_of_glow"),
    Float("particle_generation_frequency",
           SIDETIP="Hz", UNIT_SCALE=per_sec_unit_scale),
    float_sec("lifetime_of_trailing_particles"),
    float_wu_sec("velocity_of_trailing_particles"),
    float_sec("trailing_particle_min_time"),
    float_sec("trailing_particle_max_time"),

    Pad(52),
    dependency("texture",  "bitm"),
    SIZE=340,
    )

def get():
    return glw__def

glw__def = TagDef("glw!",
    blam_header('glw!'),
    glw__body,

    ext=".glow", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

class StringIdManager:
    strings = ()
    set_offsets = ()

    def __init__(self, strings, sets):
        self.strings = strings
        self.set_offsets = tuple((s[0], s[1]) for s in sets)

    def get_string(self, string_id_block):
        string_id = string_id_block.string_id
        idx_bit_ct = string_id_block.STRINGID_IDX_BITS
        set_bit_ct = string_id_block.STRINGID_SET_BITS

        set_id = (string_id >> idx_bit_ct) & ((1 << set_bit_ct) - 1)
        index = string_id & ((1 << idx_bit_ct) - 1)
        if set_id in range(len(self.set_offsets)):
            set_offset = self.set_offsets[set_id]
            if index < set_offset[0]:
                set_id -= 1
                set_offset = self.set_offsets[set_id]

            if set_id >= 0:
                index += set_offset[1] - set_offset[0]

        if index not in range(len(self.strings)):
            return ""

        return self.strings[index].string



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from ...hek.defs.objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

text_value_pair = Struct("text_value_pair",
    Bool8("flags",
        "default_setting",
        "unchanged_setting",
        ),
    Pad(3),

    SInt32("integer"),
    Float("real"),
    dependency_os("string_id", "sidy"),
    Pad(4),
    dependency_os("label_string_id", "sidy"),
    Pad(4),
    dependency_os("description_string_id", "sidy"),

    SIZE=72
    )

sily_body = Struct("tagdata",
    dependency_os("parameter", "sidy"),
    Pad(4),
    dependency_os("title_text", "sidy"),
    Pad(4),
    dependency_os("description_text", "sidy"),

    Pad(4),
    SEnum16("type",
        "integer",
        "string_id",
        "unused",
        "real",
        ),

    Pad(2),
    reflexive("text_value_pairs", text_value_pair, 32),
    SIZE=76
    )

def get():
    return sily_def

sily_def = TagDef("sily",
    blam_header_os('sily', 0),
    sily_body,

    ext=".text_value_pair_definition", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: -DeToX-
# 	Fixed some Reflexives..
# revision: 3		author: -DeToX-
# 	Named some values..
# revision: 4		author: Lord Zedd
# 	Enums, better names, etc.
# revision: 5		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

effe_event_part_camera_mode = (
    "independent_of_camera_mode",
    "first_person_only",
    "third_person_only",
    "both_first_and_third",
    )

effe_event_part_create_in_disposition = (
    "either_mode",
    "violent_mode_only",
    "nonviolent_mode_only",
    )

effe_event_part_create_in_environment = (
    "any_environment",
    "air_only",
    "water_only",
    "space_only",
    )

effe_event_particle_system_coordinate_system = (
    "world",
    "local",
    "ancestor",
    )


effe_location = Struct("location",
    h3_string_id("marker_name"),
    SInt32("unknown", VISIBLE=False),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=12
    )


effe_event_part = Struct("part",
    SEnum16("create_in_environment", *effe_event_part_create_in_environment),
    SEnum16("create_in_disposition", *effe_event_part_create_in_disposition),
    SInt16("location_index"),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    SEnum8("camera_mode", *effe_event_part_camera_mode),
    StrLatin1("anticipated_tag_class", SIZE=4),
    h3_dependency("spawned_tag"),
    QStruct("velocity_bounds", INCLUDE=from_to),
    BytesRaw("unknown_3", SIZE=8, VISIBLE=False),
    float_rad("velocity_cone_angle"),
    from_to_rad("angular_velocity_bounds"),
    QStruct("radius_modifier_bounds", INCLUDE=from_to),
    QStruct("origin_offset", INCLUDE=xyz_float),
    yp_float_rad("origin_rotation"),
    Bool32("a_scales_values", *unknown_flags_32),
    Bool32("b_scales_values", *unknown_flags_32),
    ENDIAN=">", SIZE=96
    )


effe_event_acceleration = Struct("acceleration",
    SEnum16("create_in_environment", *effe_event_part_create_in_environment),
    SEnum16("create_in_disposition", *effe_event_part_create_in_disposition),
    SInt16("location_index"),
    SInt16("unknown", VISIBLE=False),
    Float("acceleration"),
    Float("inner_cone_angle"),
    Float("outer_cone_angle"),
    ENDIAN=">", SIZE=20
    )


effe_event_particle_system_emitter_unknown_21_unknown_1 = Struct("unknown_1",
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    SInt8("input"),
    SInt8("input_range"),
    SEnum8("output_kind", *cntl_contrail_system_output_kind_0),
    SInt8("output"),
    h3_rawdata_ref("unknown_1"),
    BytesRaw("unknown_2", SIZE=8, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=36
    )


effe_event_particle_system_emitter_unknown_21 = Struct("unknown_21",
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    h3_reflexive("unknown_1", effe_event_particle_system_emitter_unknown_21_unknown_1),
    BytesRaw("unknown_2", SIZE=8, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=24
    )


effe_event_particle_system_emitter_unknown_39 = Struct("unknown_39",
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=4, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


effe_event_particle_system_emitter_compiled_function = Struct("compiled_function",
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=16, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=64
    )


effe_event_particle_system_emitter_compiled_color_function = Struct("compiled_color_function",
    color_rgb_float("color"),
    Float("magnitude"),
    ENDIAN=">", SIZE=16
    )


effe_event_particle_system_emitter = Struct("emitter",
    h3_string_id("name"),
    Bool16("unknown_0", *unknown_flags_32),
    SInt16("unknown_1", VISIBLE=False),
    BytesRaw("unknown_2", SIZE=16, VISIBLE=False),
    SInt8("input_0"),
    SInt8("input_range_0"),
    SEnum8("output_kind_0", *cntl_contrail_system_output_kind_0),
    SInt8("output_0"),
    h3_rawdata_ref("unknown_3"),
    BytesRaw("unknown_4", SIZE=32, VISIBLE=False),
    SInt8("input_1"),
    SInt8("input_range_1"),
    SEnum8("output_kind_1", *cntl_contrail_system_output_kind_0),
    SInt8("output_1"),
    h3_rawdata_ref("unknown_5"),
    BytesRaw("unknown_6", SIZE=32, VISIBLE=False),
    SInt8("input_2"),
    SInt8("input_range_2"),
    SEnum8("output_kind_2", *cntl_contrail_system_output_kind_0),
    SInt8("output_2"),
    h3_rawdata_ref("unknown_7"),
    BytesRaw("unknown_8", SIZE=8, VISIBLE=False),
    SInt8("input_3"),
    SInt8("input_range_3"),
    SEnum8("output_kind_3", *cntl_contrail_system_output_kind_0),
    SInt8("output_3"),
    h3_rawdata_ref("unknown_9"),
    BytesRaw("unknown_10", SIZE=8, VISIBLE=False),
    SInt8("input_4"),
    SInt8("input_range_4"),
    SEnum8("output_kind_4", *cntl_contrail_system_output_kind_0),
    SInt8("output_4"),
    h3_rawdata_ref("unknown_11"),
    BytesRaw("unknown_12", SIZE=8, VISIBLE=False),
    SInt8("input_5"),
    SInt8("input_range_5"),
    SEnum8("output_kind_5", *cntl_contrail_system_output_kind_0),
    SInt8("output_5"),
    h3_rawdata_ref("unknown_13"),
    BytesRaw("unknown_14", SIZE=8, VISIBLE=False),
    SInt8("input_6"),
    SInt8("input_range_6"),
    SEnum8("output_kind_6", *cntl_contrail_system_output_kind_0),
    SInt8("output_6"),
    h3_rawdata_ref("unknown_15"),
    BytesRaw("unknown_16", SIZE=8, VISIBLE=False),
    SInt8("input_7"),
    SInt8("input_range_7"),
    SEnum8("output_kind_7", *cntl_contrail_system_output_kind_0),
    SInt8("output_7"),
    h3_rawdata_ref("unknown_17"),
    BytesRaw("unknown_18", SIZE=8, VISIBLE=False),
    SInt8("input_8"),
    SInt8("input_range_8"),
    SEnum8("output_kind_8", *cntl_contrail_system_output_kind_0),
    SInt8("output_8"),
    h3_rawdata_ref("unknown_19"),
    BytesRaw("unknown_20", SIZE=8, VISIBLE=False),
    h3_dependency("particle_physics"),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    h3_reflexive("unknown_21", effe_event_particle_system_emitter_unknown_21),
    SInt8("input_9"),
    SInt8("input_range_9"),
    SEnum8("output_kind_9", *cntl_contrail_system_output_kind_0),
    SInt8("output_9"),
    h3_rawdata_ref("unknown_22"),
    BytesRaw("unknown_23", SIZE=32, VISIBLE=False),
    SInt8("input_10"),
    SInt8("input_range_10"),
    SEnum8("output_kind_10", *cntl_contrail_system_output_kind_0),
    SInt8("output_10"),
    h3_rawdata_ref("unknown_24"),
    BytesRaw("unknown_25", SIZE=8, VISIBLE=False),
    SInt8("input_11"),
    SInt8("input_range_11"),
    SEnum8("output_kind_11", *cntl_contrail_system_output_kind_0),
    SInt8("output_11"),
    h3_rawdata_ref("unknown_26"),
    BytesRaw("unknown_27", SIZE=8, VISIBLE=False),
    SInt8("input_12"),
    SInt8("input_range_12"),
    SEnum8("output_kind_12", *cntl_contrail_system_output_kind_0),
    SInt8("output_12"),
    h3_rawdata_ref("unknown_28"),
    BytesRaw("unknown_29", SIZE=8, VISIBLE=False),
    SInt8("input_13"),
    SInt8("input_range_13"),
    SEnum8("output_kind_13", *cntl_contrail_system_output_kind_0),
    SInt8("output_13"),
    h3_rawdata_ref("unknown_30"),
    BytesRaw("unknown_31", SIZE=8, VISIBLE=False),
    SInt8("input_14"),
    SInt8("input_range_14"),
    SEnum8("output_kind_14", *cntl_contrail_system_output_kind_0),
    SInt8("output_14"),
    h3_rawdata_ref("particle_scale"),
    BytesRaw("unknown_32", SIZE=8, VISIBLE=False),
    SInt8("input_15"),
    SInt8("input_range_15"),
    SEnum8("output_kind_15", *cntl_contrail_system_output_kind_0),
    SInt8("output_15"),
    h3_rawdata_ref("particle_tint"),
    BytesRaw("unknown_33", SIZE=8, VISIBLE=False),
    SInt8("input_16"),
    SInt8("input_range_16"),
    SEnum8("output_kind_16", *cntl_contrail_system_output_kind_0),
    SInt8("output_16"),
    h3_rawdata_ref("particle_alpha"),
    BytesRaw("unknown_34", SIZE=8, VISIBLE=False),
    SInt8("input_17"),
    SInt8("input_range_17"),
    SEnum8("output_kind_17", *cntl_contrail_system_output_kind_0),
    SInt8("output_17"),
    h3_rawdata_ref("particle_alpha_black_point"),
    BytesRaw("unknown_35", SIZE=8, VISIBLE=False),
    SInt32("unknown_36", VISIBLE=False),
    SInt32("unknown_37", VISIBLE=False),
    SInt32("unknown_38", VISIBLE=False),
    h3_reflexive("unknown_39", effe_event_particle_system_emitter_unknown_39),
    h3_reflexive("compiled_functions", effe_event_particle_system_emitter_compiled_function),
    h3_reflexive("compiled_color_functions", effe_event_particle_system_emitter_compiled_color_function),
    ENDIAN=">", SIZE=752
    )


effe_event_particle_system = Struct("particle_system",
    Array("unknown_array_0", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    h3_dependency("particle"),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("location_index"),
    SEnum16("coordinate_system", *effe_event_particle_system_coordinate_system),
    SEnum16("environment", *effe_event_part_create_in_environment),
    SEnum16("disposition", *effe_event_part_create_in_disposition),
    SEnum16("camera_mode", *effe_event_part_camera_mode),
    SInt16("sort_bias"),
    Bool16("flags", *unknown_flags_16),
    Array("unknown_array_1", SUB_STRUCT=Float("unknown"), SIZE=4, VISIBLE=False),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    Float("amount_size"),
    Float("unknown_3", VISIBLE=False),
    Float("lod_in_distance"),
    Float("lod_feather_in_delta"),
    h3_reflexive("emitters", effe_event_particle_system_emitter),
    Float("unknown_4", VISIBLE=False),
    ENDIAN=">", SIZE=92
    )


effe_event = Struct("event",
    h3_string_id("name"),
    SInt32("unknown", VISIBLE=False),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    Float("skip_fraction"),
    QStruct("delay_bounds", INCLUDE=from_to),
    QStruct("duration_bounds", INCLUDE=from_to),
    h3_reflexive("parts", effe_event_part),
    h3_reflexive("accelerations", effe_event_acceleration),
    h3_reflexive("particle_systems", effe_event_particle_system),
    ENDIAN=">", SIZE=68
    )


effe_unknown_9 = Struct("unknown_9",
    BytesRaw("unknown", SIZE=12, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=12
    )


effe_body = Struct("tagdata",
    Bool32("flags",
        ("dark_casings", 1 << 10),
        ),
    SInt32("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    BytesRaw("unknown_2", SIZE=4, VISIBLE=False),
    Float("unknown_3", VISIBLE=False),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    SInt16("loop_start_event"),
    SInt16("unknown_4", VISIBLE=False),
    BytesRaw("unknown_5", SIZE=4, VISIBLE=False),
    h3_reflexive("locations", effe_location),
    h3_reflexive("events", effe_event),
    h3_dependency("looping_sound"),
    SInt8("location_index"),
    SInt8("event_index"),
    SInt16("unknown_6", VISIBLE=False),
    Float("always_play_distance"),
    Float("never_play_distance"),
    Float("unknown_7", VISIBLE=False),
    Float("unknown_8", VISIBLE=False),
    h3_reflexive("unknown_9", effe_unknown_9),
    ENDIAN=">", SIZE=104
    )


def get():
    return effe_def

effe_def = TagDef("effe",
    h3_blam_header('effe'),
    effe_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["effe"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

# ########################################################################
# The element order in all the enumerators is important(DONT SHUFFLE THEM)
# ########################################################################

#Shared Enumerator options
materials_list = (
    "dirt",
    "sand",
    "stone",
    "snow",
    "wood",
    "metal_hollow",
    "metal_thin",
    "metal_thick",
    "rubber",
    "glass",
    "force_field",
    "grunt",
    "hunter_armor",
    "hunter_skin",
    "elite",
    "jackal",
    "jackal_energy_shield",
    "engineer_skin",
    "engineer_force_field",
    "flood_combat_form",
    "flood_carrier_form",
    "cyborg_armor",
    "cyborg_energy_shield",
    "human_armor",
    "human_skin",
    "sentinel",
    "moniter",
    "plastic",
    "water",
    "leaves",
    "elite_energy_shield",
    "ice",
    "hunter_shield",
    )
material_effect_types = (
    "walk",
    "run",
    "sliding",
    "shuffle",
    "jump",
    "jump_land",
    "biped_unused1",
    "biped_unused2",
    "impact",
    "vehicle_tire_slip",
    "vehicle_chassis_slip",
    "vehicle_unused1",
    "vehicle_unused2"
    )
shader_types = (
    "shdr",  # Shader
    "unused0",
    "unused1",
    "unused2",
    "senv",  # shader_environment
    "soso",  # shader_model
    "sotr",  # shader_transparent_generic
    "schi",  # shader_transparent_chicago
    "scex",  # shader_transparent_chicago_extended
    "swat",  # shader_water
    "sgla",  # shader_glass
    "smet",  # shader_meter
    "spla",  # shader_plasma
    )
object_types = (
    "obje",  # object
    "bipd",  # biped
    "vehi",  # vehicle
    "weap",  # weapon
    "eqip",  # equipment
    "garb",  # garbage
    "proj",  # projectile
    "scen",  # scenery
    "mach",  # device_machine
    "ctrl",  # device_control
    "lifi",  # device_light_fixture
    "plac",  # placeholder
    "ssce",  # sound_scenery
    )
actor_types = (
    "elite",
    "jackal",
    "grunt",
    "hunter",
    "engineer",
    "assassin",
    "player",
    "marine",
    "crew",
    "combat_form",
    "infection_form",
    "carrier_form",
    "moniter",
    "sentinal",
    "none",
    "mounted_weapon"
    )
animation_functions = (
    "one",
    "zero",
    "cosine",
    "cosine_variable_period",
    "diagonal_wave",
    "diagonal_wave_variable_period",
    "slide",
    "slide_variable_period",
    "noise",
    "jitter",
    "wander",
    "spark",
    )
sound_volumes = (
    "silent",
    "medium",
    "loud",
    "shout",
    "quiet"
    )
damage_category = (
    "none",
    "falling",
    "bullet",
    "grenade",
    "high_explosive",
    "sniper",
    "melee",
    "flame",
    "mounted_weapon",
    "vehicle",
    "plasma",
    "needle",
    "shotgun",
    )
grenade_types = (
    'human_frag',
    'covenant_plasma'
    )
fade_functions = (
    "linear",
    "early",
    "very_early",
    "late",
    "very_late",
    "cosine",
    )
detail_map_functions = (
    "double_biased_multiply",
    "multiply",
    "double_biased_add",
    )
device_functions = (
    "none",
    "power",
    "change_in_power",
    "position",
    "change_in_position",
    "locked",
    "delay",
    )
render_anchor = (
    "with_primary",
    "with_screen_space",
    "with_zsprite"
    )
render_fade_mode = (
    "none",
    "fade_when_perpendicular",
    "fade_when_parallel",
    )
render_mode = (
    "screen_facing",
    "parallel_to_direction",
    "perpendicular_to_direction"
    )
shader_flags = (
    "sort_bias",
    "nonlinear_tint",
    "dont_overdraw_fp_weapon"
    )
blend_flags = (
    "blend_in_hsv",
    "more_colors"
    )
hud_scaling_flags = (
    "dont_scale_offset",
    "dont_scale_size",
    "use_high_res_scale"
    )
hud_flash_flags = (
    "reverse_default_and_flashing_colors",
    )
hud_anchors = (
    "top_left",
    "top_right",
    "bottom_left",
    "bottom_right",
    "center"
    )
hud_panel_meter_flags = (
    "use_min_max_for_state_changes",
    "interpolate_between_min_max_flash_colors",
    "interpolate_color_along_hsv_space",
    "more_colors_for_hsv_interpolation ",
    "invert_interpolation"
    )
multitex_anchors = (
    "texture",
    "screen",
    )
multitex_wrap_modes = (
    "clamp",
    "wrap",
    )
blending_funcs = (
    "add",
    "subtract",
    "multiply",
    "multiply_2x",
    "dot",
    )
blend_functions = (
    "current",
    "next_map",
    "multiply",
    "double_multiply",
    "add",
    "add_signed_current",
    "add_signed_next_map",
    "subtract_signed_current",
    "subtract_signed_next_map",
    "blend_current_alpha",
    "blend_current_alpha_inverse",
    "blend_next_map_alpha",
    "blend_next_map_alpha_inverse",
    )
framebuffer_blend_functions = (
    "alpha_blend",
    "multiply",
    "double_multiply",
    "add",
    "subtract",
    "component_min",
    "component_max",
    "alpha_multiply_add",
    )

hmt_icon_types = (
    "a-button",
    "b-button",
    "x-button",
    "y-button",
    "black-button",
    "white-button",
    "left-trigger",
    "right-trigger",
    "dpad-up",
    "dpad-down",
    "dpad-left",
    "dpad-right",
    "start-button",
    "back-button",
    "left-thumb",
    "right-thumb",
    "left-stick",
    "right-stick",
    "action",
    "throw-grenade",
    "primary-trigger",
    "integrated-light",
    "jump",
    "use-equipment",
    "rotate-weapons",
    "rotate-grenades",
    "zoom",
    "crouch",
    "accept",
    "back",
    "move",
    "look",
    "custom-1",
    "custom-2",
    "custom-3",
    "custom-4",
    "custom-5",
    "custom-6",
    "custom-7",
    "custom-8",
    )
# DO NOT MODIFY ANY OF THESE SCRIPT ENUMS.
# The exact lettering is important as the script extractor uses
# these strings to reconstruct scripts
script_types = (
    "startup",
    "dormant",
    "continuous",
    "static",
    "stub",
    )
script_object_types = (
    "unparsed",
    "special form",
    # built-in function type. function seems to
    # be specified by the index_union value, and
    # data value seems to always be 0
    "function name",
    "passthrough",
    "void",
    "boolean",
    "real",
    "short",
    "long", # 8

# DO NOT MODIFY ANY OF THESE SCRIPT ENUMS!
# The exact lettering is important as the script extractor
# uses these strings to reconstruct scripts

    "string",
    # lower 16bits of data points to an entry in scripts
    "script", # 10

    # lower 16bits of data points to an entry in their
    # respectively named arrays in the scenario. also,
    # cutscene_recording refers to recorded_animations.
    "trigger_volume",
    "cutscene_flag",
    "cutscene_camera_point",
    "cutscene_title",
    "cutscene_recording", # 15

    # lower 16bits of data points to an entry in device groups
    "device_group",
    # lower 16bits of data points to an entry in
    # encounters array, next 8bits points to entry
    # in squad array, and upper 8bits seem to be flags,
    # with the 8th bit signifying whether or not to
    # look in the squads array of the encounter.
    # specifies a squad if set, else an encounter.
    "ai",
    # lower 16bits of data points to entry in their
    # respectively named arrays in the scenario.
    "ai_command_list",
    "starting_profile",
    "conversation", # 20

    "navpoint",
    # lower 16bits of data points to entry in
    # messages array in scenario hud_messages
    "hud_message",
    "object_list", # 23

    # lower 16bits of data in all of these
    # points to a tag in the maps tag index
    "sound",
    "effect",
    "damage",
    "looping_sound",
    "animation_graph",
    "actor_variant",
    "damage_effect",
    "object_definition", # 31

    "game_difficulty",
    "team",
    "ai_default_state",
    # lower 16bits of data points to entry in bipeds palette
    # (yes, really, not the actors palette)
    "actor_type",
    "hud_corner", # 36

    # lower 16bits of data in all of these
    # points to an entry in object names array
    "object",
    "unit",
    "vehicle",
    "weapon",
    "device",
    "scenery", # 42

    # lower 16bits of data in all of these
    # points to an entry in object names array
    "object_name",
    "unit_name",
    "vehicle_name",
    "weapon_name",
    "device_name",
    "scenery_name", # 48
    )
# DO NOT MODIFY ANY OF THESE ENUMS!
# The exact lettering is important!
# INCOMPLETE: Entries with None haven't been determined yet.
# Additionally, some are duplicates. Not sure why, but they are
#    * == taken from script bible
#   ** == educated guess based on neighboring commands
script_built_in_functions = (
    "begin", # 0
    "begin_random",
    "if",
    "cond", # *
    "set",
    "and",
    "or",
    "+",
    "-",
    "*",
    "/", # 10
    "min", # *
    "max",
    "=",
    "!=",
    ">",
    "<",
    ">=",
    "<=",
    "sleep",
    "sleep_until", # 20
    "wake",
    "inspect",
    "unit",
    "ai_debug_communication_supress", # *
    "ai_debug_communication_ignore", # *
    "ai_debug_communication_focus", # *
    "not",
    "print",
    "players",
    "volume_teleport_players_not_inside", # 30
    "volume_test_object",
    "volume_test_objects",
    "volume_test_objects_all",
    "object_teleport",
    "object_set_facing",
    "object_set_shield",
    "object_set_permutation",
    "object_create",
    "object_destroy",
    "object_create_anew", # 40
    "object_create_containing",
    "object_create_anew_containing",
    "object_destroy_containing",
    "object_destroy_all",
    "list_get",
    "list_count",
    "effect_new",
    "effect_new_on_object_marker",
    "damage_new",
    "damage_object", # 50
    "objects_can_see_object",
    "objects_can_see_flag",
    "objects_delete_by_definition",
    None,
    # these next 5 might be shifted. a little
    "sound_set_gain", # *
    "sound_set_gain", # * same as previous, but extra arg
    "script_recompile", # *
    "help", # *
    "random_range", # *
    "real_random_range", # 60
    "numeric_countdown_timer_set",
    "numeric_countdown_timer_get", # *
    "numeric_countdown_timer_stop",
    "numeric_countdown_timer_restart", # *
    "breakable_surfaces_enable",
    "recording_play",
    "recording_play_and_delete",
    "recording_play_and_hover",
    "recording_kill",
    "recording_time", # 70
    "object_set_ranged_attack_inhibited",
    "object_set_melee_attack_inhibited",
    "objects_dump_memory", # *
    "object_set_collideable",
    "object_set_scale",
    "objects_attach",
    "objects_detach",
    "garbage_collect_now",
    "object_cannot_take_damage",
    "object_can_take_damage", # 80
    "object_beautify",
    "objects_predict",
    "object_type_predict",
    "object_pvs_activate",
    "object_pvs_set_object",
    "object_pvs_set_camera",
    "object_pvs_clear",
    "render_lights",
    "scenery_get_animation_time", # *
    "scenery_animation_start", # 90
    "scenery_animation_start_at_frame",
    "render_effects", # *
    "unit_can_blink", # *
    "unit_open",
    "unit_close",
    "unit_kill",
    "unit_kill_silent", # *
    "unit_get_custom_animation_time",
    "unit_stop_custom_animation",
    "unit_custom_animation_at_frame", # 100
    "custom_animation",
    "custom_animation_list",
    "unit_is_playing_custom_animation",
    "unit_aim_without_turning", # *
    "unit_set_emotion",
    "unit_set_enterable_by_player",
    "unit_enter_vehicle",
    "vehicle_test_seat_list",
    "vehicle_test_seat",
    "unit_set_emotion_animation", # * 110
    "unit_exit_vehicle",
    "unit_set_maximum_vitality",
    "units_set_maximum_vitality", # *
    "unit_set_current_vitality",
    "units_set_current_vitality",
    "vehicle_load_magic",
    "vehicle_unload",
    "magic_seat_name", # *
    "unit_set_seat",
    "magic_melee_attack", # * 120
    "vehicle_riders",
    "vehicle_driver", # *
    "vehicle_gunner",
    "unit_get_health",
    "unit_get_shield",
    "unit_get_total_grenade_count",
    "unit_has_weapon",
    "unit_has_weapon_readied",
    "unit_doesnt_drop_items",
    "unit_impervious", # 130
    "unit_suspended",
    "unit_solo_player_integrated_night_vision_is_active",
    "units_set_desired_flashlight_state",
    "unit_set_desired_flashlight_state",
    "unit_get_current_flashlight_state",
    "device_set_never_appears_locked",
    "device_get_power",
    "device_set_power",
    "device_set_position",
    "device_get_position", # 140
    "device_set_position_immediate",
    "device_group_get",
    "device_group_set",
    "device_group_set_immediate",
    "device_one_sided_set",
    "device_operates_automatically_set",
    "device_group_change_only_once_more_set",
    "breakable_surfaces_reset",
    "cheat_all_powerups", # *
    "cheat_all_weapons", # * 150
    "cheat_spawn_warthog", # *
    "cheat_all_vehicles", # *
    "cheat_teleport_to_camera", # *
    "cheat_active_camouflage", # *
    # no space for this command. Might not actually exist
    #"cheat_active_camouflage_local_player", # *
    "cheats_load", # *
    "ai_free",
    "ai_free_units",
    "ai_attach",
    "ai_attach",
    "ai_detach", # 160
    "ai_place",
    "ai_place",
    "ai_kill",
    "ai_kill_silent",
    "ai_erase",
    "ai_erase_all",
    "ai_select", # *
    # no space for this command. Might not actually exist
    #"ai_deselect", # *
    "ai_spawn_actor",
    "ai_spawn_actor",
    "ai_set_deaf", # 170
    "ai_set_blind",
    "ai_magically_see_encounter",
    "ai_magically_see_encounter",
    "ai_magically_see_players",
    "ai_magically_see_unit", # *
    "ai_timer_expire",
    "ai_attack",
    "ai_attack",
    "ai_retreat",
    "ai_maneuver", # 180
    "ai_maneuver",
    "ai_migrate",
    "ai_migrate",
    "ai_migrate_and_speak",
    "ai_migrate_by_unit",
    "ai_allegiance",
    "ai_living_count",
    "ai_living_count",
    "ai_strength",
    "ai_strength", # 190
    "ai_nonswarm_count",
    "ai_nonswarm_count",
    "ai_actors",
    "ai_go_to_vehicle",
    "ai_going_to_vehicle",
    "ai_exit_vehicle",
    "ai_braindead",
    "ai_braindead",
    "ai_braindead_by_unit",
    "ai_prefer_target", # 200
    "ai_teleport_to_starting_location",
    "ai_teleport_to_starting_location_if_unsupported",
    "ai_teleport_to_starting_location_if_unsupported",
    "ai_renew",
    "ai_try_to_fight",
    "ai_try_to_fight",
    "ai_try_to_fight_player",
    "ai_command_list",
    "ai_command_list_by_unit",
    "ai_command_list_advance", # 210
    "ai_command_list_status",
    "ai_command_list_status",
    "ai_force_active",
    "ai_force_active",
    "ai_set_return_state", # *
    "ai_set_current_state", # *
    "ai_playfight",
    "ai_playfight",
    "ai_status",
    "ai_vehicle_encounter", # 220
    "ai_vehicle_enterable_distance",
    "ai_vehicle_enterable_distance",
    "ai_vehicle_enterable_actor_type",
    "ai_vehicle_enterable_actors", # *
    "ai_vehicle_enterable_disable",
    "ai_look_at_object", # *
    "ai_stop_looking", # *
    "ai_automatic_migration_target",
    "ai_automatic_migration_target",
    "ai_follow_target_disable", # 230
    "ai_follow_target_players",
    "ai_follow_target_ai",
    "ai_follow_distance",
    "ai_conversation",
    "ai_conversation",
    "ai_conversation_stop",
    "ai_conversation_advance",
    "ai_conversation_status",
    "ai_conversation_status",
    "ai_link_activation", # 240
    "ai_berserk",
    "ai_set_team",
    "ai_allow_dormant",
    "ai_allow_dormant",
    "ai_allegiance_broken",
    "camera_control",
    "camera_set",
    "camera_set_relative",
    "camera_set_first_person",
    "camera_set_dead", # 250
    "camera_set_dead",
    "camera_time",
    "debug_camera_load",
    "debug_camera_save",
    "game_speed",
    "game_variant", # *
    "game_difficulty_get",
    "game_difficulty_get",
    "game_difficulty_get_real",
    "profile_service_clear_timers", # * 260
    "profile_service_dump_timers", # *
    "map_reset",
    "map_name",
    "switch_bsp",
    "structure_bsp_index",
    "crash", # *
    "switch_bsp",
    "structure_bsp_index",
    "version", # *
    "playback", # * 270
    "quit", # *
    "texture_cache_flush", # *
    "sound_cache_flush", # *
    "sound_cache_dump_to_file", # *
    "debug_memory", # *
    "debug_memory_by_file", # *
    "debug_memory_for_file", # *
    "debug_tags", # *
    "profile_reset", # *
    "profile_dump", # *280
    # no space for these commands. Might not actually exist
    #"profile_activate", # *
    #"profile_deactivate", # *
    #"profile_graph_toggle", # *
    "debug_pvs", # *
    "radiosity_start", # *
    # no space for these commands. Might not actually exist
    #"radiosity_save", # *
    #"radiosity_debug_point", # *
    "ai",
    "ai_dialogue_triggers",
    "ai_grenades",
    None, None,
    "ai",
    "ai_dialogue_triggers",
    "ai_grenades", # 290
    None,
    "fade_in",
    "fade_out",
    "cinematic_start",
    "cinematic_stop",
    "cinematic_skip_start_internal",
    "fade_in",
    "fade_out",
    "cinematic_start",
    "cinematic_stop", # 300
    "cinematic_suppress_bsp_object_creation",
    "cinematic_skip_start_internal",
    "cinematic_skip_stop_internal",
    "cinematic_show_letterbox",
    "cinematic_set_title",
    "game_all_quiet",
    "game_safe_to_speak",
    "game_is_cooperative",
    "game_won",
    "game_lost", # 310
    "game_safe_to_save",
    "game_save_totally_unsafe",
    "game_saving",
    "game_is_cooperative",
    "game_reverted",
    "game_save_cancel",
    "game_save_no_timeout",
    "game_save_totally_unsafe",
    "game_saving",
    "game_revert", # 320
    "game_reverted",
    "game_skip_ticks",
    "sound_impulse_start",
    "sound_impulse_time",
    "sound_impulse_stop",
    "sound_looping_start",
    "sound_looping_start",
    "game_skip_ticks",
    "sound_looping_set_alternate",
    "sound_impulse_start", # 330
    "sound_impulse_time",
    "sound_impulse_stop",
    "sound_looping_predict", # *
    "sound_looping_start",
    "sound_looping_stop",
    "vehicle_hover",
    "sound_looping_set_alternate",
    "player_enable_input",
    "player_camera_control",
    "player_action_test_reset", # 340
    "player_action_test_jump", # *
    "player_action_test_primary_trigger",
    "player_action_test_grenade_trigger",
    "player_action_test_zoom", # *
    "player_action_test_action",
    "player_action_test_accept",
    "player_action_test_back",
    "sound_class_set_gain",
    "vehicle_hover",
    "players_unzoom_all", # 350
    "player_enable_input",
    "player_action_test_look_relative_all_directions",
    "player_action_test_reset",
    "player_add_equipment",
    "show_hud",
    "show_hud",
    "show_hud_help_text",
    "enable_hud_help_flash",
    "hud_help_flash_restart", # *
    "activate_nav_point_flag", # *360
    # no space for this command. Might not actually exist
    #"activate_nav_point_object", # *
    "activate_team_nav_point_flag",
    "activate_team_nav_point_flag",
    "activate_team_nav_point_object",
    "deactivate_nav_point_flag", # *
    # no space for this command. Might not actually exist
    #"deactivate_nav_point_object", # *
    "deactivate_team_nav_point_flag",
    "deactivate_team_nav_point_flag",
    "deactivate_team_nav_point_object",
    "cls",
    "show_hud",
    "show_hud_help_text", # 370
    "player_effect_set_max_translation",
    "player_effect_set_max_rotation",
    "player_effect_set_max_rumble",
    "player_effect_start",
    "activate_team_nav_point_flag",
    "activate_team_nav_point_object",
    "hud_blink_health",
    "hud_show_shield",
    "deactivate_team_nav_point_flag",
    "deactivate_team_nav_point_object", # 380
    "hud_blink_motion_sensor",
    "hud_show_crosshair",
    "hud_set_help_text",
    "hud_set_help_text",
    "hud_set_objective_text",
    "hud_set_timer_time",
    "hud_set_timer_warning_time",
    "hud_set_timer_position",
    "show_hud_timer",
    "pause_hud_timer", # 390
    "hud_get_timer_ticks",
    "time_code_show",
    "time_code_start",
    "time_code_reset", # *
    None, None,
    "rasterizer_model_ambient_reflection_tint",
    "rasterizer_lights_reset_for_new_map",
    "cinematic_screen_effect_start",
    "player_effect_set_max_translation", # 400
    "player_effect_set_max_rotation",
    "player_effect_set_max_rumble",
    "player_effect_start",
    "player_effect_stop",
    "cinematic_screen_effect_stop",
    "cinematic_set_near_clip_distance",
    "hud_show_shield",
    "hud_show_health", # **
    "hud_show_motion_sensor",
    "hud_show_crosshair", # ** 410
    "player0_look_invert_pitch",
    "player0_look_pitch_is_inverted",
    "hud_set_help_text",
    "hud_set_objective_text",
    "display_scenario_help",
    None, None, None, None,
    None, # 420
    None, None, None, None, None, None, None,
    "rasterizer_decals_flush",
    "rasterizer_fps_accumulate", # *
    "rasterizer_model_ambient_reflection_tint", # 430
    "rasterizer_lights_reset_for_new_map", # *
    "script_screen_effect_set_value", # *
    "cinematic_screen_effect_start",
    "cinematic_screen_effect_set_convolution",
    "cinematic_screen_effect_set_filter",
    "cinematic_screen_effect_set_filter_desaturation_tint",
    "cinematic_screen_effect_set_video", # *
    "cinematic_screen_effect_stop",
    None,
    None, # 440
    None, None, None, None, None, None, None, None, None,
    None, # 450
    None, None, None, None, None, None, None, None, None,
    None, # 460
    None, None, None, None, None, None, None, None, None,
    None, # 470
    None, None, None, None, None, None, None, None, None,
    None, # 480
    None, None, None, None, None, None, None, None, None,
    None, # 490
    None, None, None, None, None, None, None, None, None,
    None, # 500
    None, None, None, None, None,
    "sv_say",
    None, None, None,
    None, # 510
    None, None, None, None, None, None, None, None, None,
    None, # 520
    None, None, None, None, None, None, None, None, None,
    None, # 530
    None, None, None, None, None, None, None, None, None,
    None, # 540
    None, None, None, None, None, None, None, None, None,
    None, # 550
    None, None,
    # OPEN SAUCE SHEET
    "objects_distance_to_object",
    "object_data_get_real",
    "object_data_set_real",
    "weapon_data_get_real",
    "weapon_data_set_real",
    "weapon_data_trigger_set_real",
    None,
    "unit_data_get_integer", # 560
    "unit_data_get_object",
    "unit_data_get_integer",
    "unit_data_set_integer",
    "unit_data_get_real",
    "unit_data_set_real",
    "physics_get_gravity", # *
    "physics_set_gravity",
    "physics_constants_reset", # *
    "physics_get_gravity", # *
    None, None,
    None, # 570
    None, None, None, None, None, None, None,
    "abs_integer", # *
    "abs_real",
    None, # 580
    "pp_get_effect_instance_index_by_name", # * 580
    "pp_set_effect_instance_active",
    "pp_set_effect_instance_fade",
    "pp_set_effect_instance_active",
    "pp_set_effect_instance_fade",
    None,
    "pp_get_effect_is_valid",
    "pp_get_effect_shader_variable_index_by_name", # *
    "pp_set_effect_shader_variable_boolean",
    "pp_set_effect_shader_variable_integer", # *
    "pp_set_effect_shader_variable_real", # * 590
    "pp_set_effect_shader_instance_active", # *
    "pp_bloom_set_size", # *
    "pp_bloom_set_exposure",
    "pp_bloom_set_mix_amount",
    "pp_bloom_set_exposure",
    "pp_bloom_set_mix_amount",
    "pp_bloom_set_minimum_color", # *
    "pp_bloom_set_maximum_color", # *
    "bitwise_and", # *
    "bitwise_or", # *600
    "bitwise_xor", # *
    "bitwise_lhs", # *
    "bitwise_rhs", # *
    "bit_test",
    "bit_toggle",
    "bit_flags_test", # *
    "bit_flags_toggle", # *
    None, None,
    None, # 610
    "runtime_vector_get_element",
    "runtime_vector_set_element",
    "runtime_vector_set",
    "runtime_vector_operation",
    "runtime_vector_to_string", # *
    "object_data_set_vector",
    "object_data_save_vector",
    "display_scripted_ui_widget",
    "play_bink_movie", # *
    "hex_string_to_long", # * 620
    None, None, None,
    "structure_bsp_set_lightmap_set",
    "structure_bsp_set_sky_set",
    "ai_transforms_enabled", # *
    "ai_transform_actors",
    "ai_transform_actors_by_type", # *
    "ai_actor_is_transforming", # *
    )
_script_built_in_functions_test = list(script_built_in_functions)
#print("Current number of valid built-in script functions: %s" %
#      sum(i is not None for i in _script_built_in_functions_test))

def TEST_PRINT_HSC_BUILT_IN_FUNCTIONS():
    i = 0
    for string in _script_built_in_functions_test:
        if string is None:
            print('    None,', end="")
        else:
            print('    "%s",' % string, end="")

        if not (i % 10):
            print(" # %s" % i, end="")
        print()
        i += 1

    print("Current number of valid built-in script functions: %s" %
          sum(i is not None for i in script_built_in_functions))


function_names = (
    "none",
    "A",
    "B",
    "C",
    "D",
    )
function_inputs_outputs = (
    "none",
    "A_in",
    "B_in",
    "C_in",
    "D_in",
    "A_out",
    "B_out",
    "C_out",
    "D_out",
    )
function_inputs = (
    "none",
    "A_in",
    "B_in",
    "C_in",
    "D_in",
    )
function_outputs = (
    "none",
    "A_out",
    "B_out",
    "C_out",
    "D_out",
    )
#Tag class specific enumerators
object_export_to = (
    'none',
    'body_vitality',
    'shield_vitality',
    'recent_body_damage',
    'recent_shield_damage',
    'random_constant',
    'umbrella_shield_vitality',
    'shield_stun',
    'recent_umbrella_shield_vitality',
    'umbrella_shield_stun',
    'region_0_damage',
    'region_1_damage',
    'region_2_damage',
    'region_3_damage',
    'region_4_damage',
    'region_5_damage',
    'region_6_damage',
    'region_7_damage',
    'alive',
    'compass',
    )
weapon_export_to = (
    'none',
    'heat',
    'primary_ammunition',
    'secondary_ammunition',
    'primary_rate_of_fire',
    'secondary_rate_of_fire',
    'ready',
    'primary_ejection_port',
    'secondary_ejection_port',
    'overheated',
    'primary_charged',
    'secondary_charged',
    'illumination',
    'age',
    'integrated_light',
    'primary_firing',
    'secondary_firing',
    'primary_firing_on',
    'secondary_firing_on',
    )
biped_inputs = (
    'none',
    'flying_velocity'
    )
projectile_inputs = (
    "none",
    "range_remaining",
    "time_remaining",
    "tracer",
    )
unit_inputs = (
    "none",
    "driver_seat_power",
    "gunner_seat_power",
    "aiming_change",
    "mouth_aperture",
    "integrated_light_power",
    "can_blink",
    "shield_sapping"
    )
unit_teams = (
    "none",
    "player",
    "human",
    "covenant",
    "flood",
    "sentinel",
    "unused6",
    "unused7",
    "unused8",
    "unused9",
    )
vehicle_inputs = (
    "none",
    "speed_absolute",
    "speed_forward",
    "speed_backward",
    "slide_absolute",
    "slide_left",
    "slide_right",
    "speed_slide_maximum",
    "turn_absolute",
    "turn_left",
    "turn_right",
    "crouch",
    "jump",
    "walk",
    "velocity_air",
    "velocity_water",
    "velocity_ground",
    "velocity_forward",
    "velocity_left",
    "velocity_up",
    "left_tread_position",
    "right_tread_position",
    "left_tread_velocity",
    "right_tread_velocity",
    "front_left_tire_position",
    "front_right_tire_position",
    "back_left_tire_position",
    "back_right_tire_position",
    "front_left_tire_velocity",
    "front_right_tire_velocity",
    "back_left_tire_velocity",
    "back_right_tire_velocity",
    "wingtip_contrail",
    "hover",
    "thrust",
    "engine_hack",
    "wingtip_contrail_new",
    )
vehicle_types = (
    "human_tank",
    "human_jeep",
    "human_boat",
    "human_plane",
    "alien_scout",
    "alien_fighter",
    "turret",
    )
weapon_types = (
    "undefined",
    "shotgun",
    "needler",
    "plasma_pistol",
    "plasma_rifle",
    )
trans_shdr_properties = (
    "alpha_tested",
    "decal",
    "two_sided",
    "first_map_is_in_screenspace",
    "draw_before_water",
    "ignore_effect",
    "scale_first_map_with_distance",
    "numeric",
    )
trans_shdr_first_map_type = (
    "map_2d",
    "reflection_cube_map",
    "object_centered_cube_map",
    "viewer_centered_cube_map",
    )
detail_mask = (
    "none",
    "reflection_mask_inverse",
    "reflection_mask",
    "self_illumination_mask_inverse",
    "self_illumination_mask",
    "color_change_mask_inverse",
    "color_change_mask",
    "auxiliary_mask_inverse",
    "auxiliary_mask"
    )

# DO NOT RENAME THESE OPTIONS.
# animation extraction/compilation depends on
# the anim_types and anim_frame_info_types
# options being named exactly like this
anim_types = (
    "base",
    "overlay",
    "replacement"
    )
anim_frame_info_types = (
    "none",
    "dx_dy",
    "dx_dy_dyaw",
    "dx_dy_dz_dyaw",
    )
unit_animation_names = (
    'airborne-dead', 'landing-dead',
    'acc-front-back', 'acc-left-right', 'acc-up-down',
    'push', 'twist', 'enter', 'exit', 'look', 'talk', 'emotions', 'unused0',
    'user0', 'user1', 'user2', 'user3', 'user4',
    'user5', 'user6', 'user7', 'user8', 'user9',
    'flying-front', 'flying-back', 'flying-left', 'flying-right',
    'opening', 'closing', 'hovering'
    )
unit_weapon_animation_names = (
    'idle', 'gesture', 'turn-left', 'turn-right',
    'dive-front', 'dive-back', 'dive-left', 'dive-right',
    'move-front', 'move-back', 'move-left', 'move-right',
    'slide-front', 'slide-back', 'slide-left', 'slide-right',
    'airborne', 'land-soft', 'land-hard', 'unused0', 'throw-grenade',
    'disarm', 'drop', 'ready', 'put-away', 'aim-still', 'aim-move',
    'surprise-front', 'surprise-back', 'berserk',
    'evade-left', 'evade-right', 'signal-move', 'signal-attack', 'warn',
    'stunned-front', 'stunned-back', 'stunned-left', 'stunned-right',
    'melee', 'celebrate', 'panic', 'melee-airborne', 'flaming',
    'resurrect-front', 'resurrect-back', 'melee-continuous',
    'feeding', 'leap-start', 'leap-airborne', 'leap-melee',
    'zapping', 'unused1', 'unused2', 'unused3'
    )
unit_weapon_type_animation_names = (
    'reload-1', 'reload-2', 'chamber-1', 'chamber-2',
    'fire-1', 'fire-2', 'charged-1', 'charged-2',
    'melee', 'overheat'
    )
vehicle_animation_names = (
    'steering', 'roll', 'throttle', 'velocity',
    'braking', 'ground-speed', 'occupied', 'unoccupied'
    )
weapon_animation_names = (
    'idle', 'ready', 'put-away',
    'reload-1', 'reload-2', 'chamber-1', 'chamber-2',
    'charged-1', 'charged-2', 'fire-1', 'fire-2'
    )
device_animation_names = (
    'position','power'
    )
fp_animation_names = (
    'idle', 'posing', 'fire-1',
    'moving', 'overlays', 'light-off', 'light-on',
    'reload-empty', 'reload-full', 'overheated', 'ready', 'put-away',
    'overcharged', 'melee', 'fire-2', 'overcharged-jitter',
    'throw-grenade', 'ammunition', 'misfire-1', 'misfire-2',
    'throw-overheated', 'overheating', 'overheating-again',
    'enter', 'exit-empty', 'exit-full', 'o-h-exit', 'o-h-s-enter'
    )
unit_damage_animation_names = []
for typ in ("s-ping", "h-ping", "s-kill", "h-kill"):
    for side in ("front", "left", "right", "back"):
        for reg in ("gut", "chest", "head",
                    "l-arm", "l-hand", "l-leg", "l-foot",
                    "r-arm", "r-hand", "r-leg", "r-foot"):
            unit_damage_animation_names.append("%s %s %s" % (typ, side, reg))

del typ
del side
del reg

# ########################################################################
# The element order in all the enumerators is important(DONT SHUFFLE THEM)
# ########################################################################

#Shared Enumerator options
grenade_types_os = (
    'human_frag',
    'covenant_plasma',
    'custom_2',
    'custom_3',
    )

actor_states = (
    'none',
    'sleeping',
    'alert',
    'moving_repeat_same_position',
    'moving_loop',
    'moving_loop_back_and_forth',
    'moving_loop_randomly',
    'moving_randomly',
    'guarding',
    'guarding_at_guard_position',
    'searching',
    'fleeing'
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.actr import *
from ..common_descs import *

actr_body = dict(actr_body)
actr_body[3] = SEnum16("type", *actor_types)
actr_body[12] = dict(actr_body[12])
actr_body[12][2] = SEnum16("leader_type", *actor_types)

actr_body[14] = dict(actr_body[14])
actr_body[14][6] = SEnum16("defensive_crouch_type",
    "never",
    "danger",
    "low_shields",
    "hide_behind_shield",
    "any_target",
    )


def get():
    return actr_def

actr_def = TagDef("actr",
    blam_header_stubbs('actr', 2),
    actr_body,

    ext=".actor", endian=">", tag_cls=ActrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = (
    "_fx_", "adlg", "ant_", "beam", "bipd", "bitm", "bkey", "bloc",
    "bmp3", "bsdt", "cddf", "cfxs", "chad", "char", "chdt", "chgd",
    "chmt", "cine", "cisc", "clwd", "cntl", "coll", "colo", "crea",
    "crte", "ctrl", "dctr", "decs", "draw", "drdf", "dsrc", "effe",
    "effg", "efsc", "eqip", "flck", "foot", "gint", "glps", "glvs",
    "goof", "grup", "hlmt", "hlsl", "jmad", "jmrq", "jpt_", "lens",
    "ligh", "lsnd", "lst3", "lswd", "ltvl", "mach", "matg", "mdl3",
    "mdlg", "mffn", "mode", "mulg", "nclt", "perf", "phmo", "pixl",
    "play", "pmdf", "pmov", "pphy", "proj", "prt3", "rasg", "rmcs",
    "rmct", "rmd_", "rmdf", "rmfl", "rmhg", "rmop", "rmsh", "rmt2",
    "rmtr", "rmw_", "rwrd", "sLdT", "sbsp", "scen", "scn3", "scnr",
    "sddt", "sefc", "sfx_", "sgp_", "shit", "sily", "skn3", "skya",
    "sncl", "snd_", "snde", "snmx", "spk_", "ssce", "styl", "term",
    "trak", "txt3", "udlg", "ugh_", "uise", "unic", "vehi", "vtsh",
    "wacd", "wclr", "weap", "wezr", "wfon", "wgan", "wgtz", "wigl",
    "wind", "wpos", "wrot", "wscl", "wspr", "wtuv", "zone",
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from supyr_struct.defs.common_descs import *
from supyr_struct.field_types import *
from supyr_struct.defs.tag_def import TagDef

from binilla.constants import VISIBILITY_METADATA

def get(): return composer_playlist_def


def get_str_len(parent=None, attr_index=None, path=None, new_value=None,
                **kwargs):
    if parent is None:
        raise KeyError()
    elif new_value is not None:
        parent.set_neighbor(path, new_value - 1)
        return
    else:
        return parent.get_neighbor(path) + 1


command = Struct("command",
    UEnum32("sig",
        ("Cmnd", "dnmC"),
        DEFAULT="dnmC", VISIBLE=VISIBILITY_METADATA,
        ),
    UInt16("perm_index"),
    Pad(2),
    BitStruct("play_info",
        UBitInt("loop_count", SIZE=10),
        Bit("endless_loop"),
        Bit("play_any"),
        Bit("play_alt"),
        Bit("play_ordered"),
        Bit("skip"),
        SIZE=4
        ),
    SIZE=16
    )


command_list = Struct("command_list",
    UEnum32("sig",
        ("Clst", "tslC"),
        DEFAULT="tslC", VISIBLE=VISIBILITY_METADATA
        ),
    UInt32("lsnd_path_len", VISIBLE=VISIBILITY_METADATA),
    UInt32("lsnd_path_pointer", VISIBLE=VISIBILITY_METADATA),

    UInt32("command_count", VISIBLE=VISIBILITY_METADATA),
    UInt32("commands_pointer", VISIBLE=VISIBILITY_METADATA),

    UInt32("lsnd_tag_pointer", VISIBLE=VISIBILITY_METADATA),
    UInt32("play_length", DEFAULT=180),

    SIZE=64,
    STEPTREE=Container("data",
        StrLatin1("lsnd_path", SIZE=lambda *a, **k:
            get_str_len(*a, path="..lsnd_path_len", **k)),
        Array("commands", SUB_STRUCT=command, SIZE="..command_count")
        )
    )


header = Struct("header",
    UEnum32("sig",
        ("Play", "yalP"),
        DEFAULT="yalP", VISIBLE=VISIBILITY_METADATA,
        ),

    UInt32("tags_dir_len", VISIBLE=VISIBILITY_METADATA),
    UInt32("tags_dir_pointer", VISIBLE=VISIBILITY_METADATA),

    UInt32("command_list_count", VISIBLE=VISIBILITY_METADATA),
    UInt32("command_list_pointer", VISIBLE=VISIBILITY_METADATA),

    Bool32("flags",
        "shuffle",
        "loop",
        ),
    SIZE=64
    )


composer_playlist_def = TagDef('composer_playlist',
    header,
    StrLatin1("tags_dir", SIZE=lambda *a, **k:
        get_str_len(*a, path=".header.tags_dir_len", **k)),
    Array("command_lists",
        SIZE=".header.command_list_count",
        SUB_STRUCT=command_list
        ),

    ext='.play', endian='<',
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = ( 'JmsMarker', )


class JmsMarker:
    __slots__ = (
        "name", "permutation",
        "region", "parent",
        "rot_i", "rot_j", "rot_k", "rot_w",
        "pos_x", "pos_y", "pos_z",
        "radius",
        )
    def __init__(self, name="", permutation="", region=0, parent=0,
                 rot_i=0.0, rot_j=0.0, rot_k=0.0, rot_w=1.0,
                 pos_x=0.0, pos_y=0.0, pos_z=0.0, radius=0.0):
        self.name = name
        self.permutation = permutation
        self.parent = parent
        self.region = max(0, region)
        self.rot_i = rot_i
        self.rot_j = rot_j
        self.rot_k = rot_k
        self.rot_w = rot_w
        self.pos_x = pos_x
        self.pos_y = pos_y
        self.pos_z = pos_z
        self.radius = radius

    def __repr__(self):
        return """JmsMarker(name=%s,
    permutation=%s,
    region=%s,  parent=%s,
    i=%s, j=%s, k=%s, w=%s,
    x=%s, y=%s, z=%s,
    radius=%s
)""" % (self.name, self.permutation, self.region, self.parent,
        self.rot_i, self.rot_j, self.rot_k, self.rot_w,
        self.pos_x, self.pos_y, self.pos_z, self.radius)

    def __eq__(self, other):
        if not isinstance(other, JmsMarker):
            return False
        elif self.name != other.name:
            return False
        elif self.permutation != other.permutation:
            return False
        elif self.region != other.region:
            return False
        elif abs(self.radius - other.radius) > 0.00001:
            return False
        elif (abs(self.rot_i - other.rot_i) > 0.00001 or
              abs(self.rot_j - other.rot_j) > 0.00001 or
              abs(self.rot_k - other.rot_k) > 0.00001 or
              abs(self.rot_w - other.rot_w) > 0.00001):
            return False
        elif (abs(self.pos_x - other.pos_x) > 0.000001 or
              abs(self.pos_y - other.pos_y) > 0.000001 or
              abs(self.pos_z - other.pos_z) > 0.000001):
            return False
        return True



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from ...hek.defs.objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

# I guess this is a tag that will be expanded upon later?

def get():
    return eqhi_def

eqhi_body = Struct("tagdata",
    SIZE=0
    )

eqhi_def = TagDef("eqhi",
    blam_header_os("eqhi", 0),
    eqhi_body,

    ext=".equipment_hud_interface", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from pathlib import Path
import os

from binilla.handler import Handler
from reclaimer.misc.defs import __all__ as all_def_names


class MiscHaloLoader(Handler):
    frozen_imp_paths = all_def_names
    default_defs_path = "reclaimer.misc.defs"

    tagsdir = tagsdir = str(Path.cwd().joinpath("tags"))

    def get_def_id(self, filepath):
        '''docstring'''
        def_id = Handler.get_def_id(self, filepath)

        if def_id == 'xbox_gametype' and 'pc_gametype' in self.id_ext_map:
            return 'pc_gametype'
        return def_id



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...os_v3_hek.defs.scnr import *

lightmap_set = Struct("lightmap_set",
    ascii_str32("name"),
    Pad(4),
    dependency_os("std_lightmap", "bitm"),
    dependency_os("dir_lightmap_1", "bitm"),
    dependency_os("dir_lightmap_2", "bitm"),
    dependency_os("dir_lightmap_3", "bitm"),
    SIZE=124
    )


sky_set_sky = Struct("sky",
    Pad(2),
    dyn_senum16("sky_index",
        DYN_NAME_PATH="...........skies.STEPTREE[DYN_I].sky.filepath"),
    dependency_os("sky", "sky "),
    SIZE=20
    )


sky_set = Struct("sky_set",
    ascii_str32("name"),
    reflexive("skies", sky_set_sky, 8),
    SIZE=44
    )


bsp_modifier = Struct("bsp_modifier",
    Pad(2),
    dyn_senum16("bsp_index",
        DYN_NAME_PATH=".....structure_bsps.STEPTREE[DYN_I].structure_bsp.filepath"),
    reflexive("lightmap_sets", lightmap_set, 64),
    reflexive("sky_sets", sky_set, 64),
    SIZE=64
    )


scnr_body = dict(scnr_body)
scnr_body[64] = reflexive("bsp_modifiers", bsp_modifier, 32)

def get():
    return scnr_def

scnr_def = TagDef("scnr",
    blam_header('scnr', 2),
    scnr_body,

    ext=".scenario", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

ngpr_body = Struct("tagdata",
    ascii_str32("name"),
    QStruct("primary_color",   INCLUDE=rgb_float),
    QStruct("secondary_color", INCLUDE=rgb_float),

    dependency("pattern", "bitm"),
    SInt16("pattern_bitmap_index"),
    Pad(2),
    dependency("decal", "bitm"),
    SInt16("decal_bitmap_index"),
    SIZE=896
    )


def get():
    return ngpr_def

ngpr_def = TagDef("ngpr",
    blam_header('ngpr', 2),
    ngpr_body,

    ext=".preferences_network_game", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from ...hek.defs.objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

build_info = Struct("build_info",
    Pad(2),
    SEnum16("build_stage",
        "ship",
        "alpha",
        "beta",
        "delta",
        "epsilon",
        "release",
        ),
    SInt32("revision"),
    SIZE=48
    )

scripted_ui_widget = Struct("scripted_ui_widget",
    ascii_str32("name"),
    dependency_os("definition", "DeLa"),
    SIZE=76
    )

parameter = Struct("parameter",
    SEnum16("type", *script_object_types),
    SIZE=2
    )

new_function = Struct("new_function",
    ascii_str32("name1"),
    ascii_str32("name2"),
    SInt16("override_index"),
    SEnum16("return_type", *script_object_types),
    reflexive("parameters", parameter, 32),
    SIZE=80
    )

new_global = Struct("new_global",
    ascii_str32("name1"),
    ascii_str32("name2"),
    SInt16("override_index"),
    SEnum16("type", *script_object_types),
    SIZE=68
    )

yelo_scripting = Struct("yelo_scripting",
    reflexive("new_functions", new_function, 464,
        DYN_NAME_PATH='.name1'),
    reflexive("new_globals", new_global, 880,
        DYN_NAME_PATH='.name1'),
    SIZE=24
    )

yelo_body = Struct("tagdata",
    SInt16("version"),
    Bool16("flags",
        "dont_fix_ui_game_globals",
        "game_updates_ignore_player_pvs_hack",
        ),
    dependency_os("yelo_globals", "gelo"),
    dependency_os("globals_override", "matg"),
    dependency_os("scenario_explicit_references", "tagc"),
    reflexive("build_info", build_info, 1),

    Pad(40),
    reflexive("scripted_ui_widgets", scripted_ui_widget, 128,
        DYN_NAME_PATH='.name'),

    Pad(16),
    # Physics
    Float("gravity_scale",      MIN=0.0, MAX=2.0, SIDETIP="[0,2]"),
    Float("player_speed_scale", MIN=0.0, MAX=6.0, SIDETIP="[0,6]"),

    Pad(44),
    Bool32("gameplay_model",
        "prohibit_multi_team_vehicles",
        ),

    Pad(20),
    reflexive("yelo_scripting", yelo_scripting, 1),
    Pad(12),#reflexive("unknown", void_desc),

    SIZE=312
    )

def get():
    return yelo_def

yelo_def = TagDef("yelo",
    blam_header_os('yelo', 2),
    yelo_body,

    ext=".project_yellow", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from supyr_struct.defs.block_def import BlockDef
from supyr_struct.field_types import QStruct, Struct, Container, Void,\
     UBitInt, BitStruct, Float, UInt16, UInt32,\
     UInt32Array, UInt16Array, FloatArray
# these structure definitions aren't really used in any code, but
# are a good way to illustrate the structure of the compressed data.
# for uncompressed animations, the data is stored as:
#     rotation first, then translation, and finally scale

keyframe_header_def = BlockDef("keyframe_header",
    UBitInt("keyframe_count", SIZE=12),
    UBitInt("keyframe_data_offset", SIZE=20),
    SIZE=4, TYPE=BitStruct
    )

# compressed quaternion rotation
# Bits AND bytes are ordered left to right as most significant to least
#   bbbbbbbb aaaaaaaa   dddddddd cccccccc   ffffffff eeeeeeee
#   iiiiiiii iiiijjjj   jjjjjjjj kkkkkkkk   kkkkwwww wwwwwwww
quaternion48_def = BlockDef("quaternion48",
    # you'll need to read the 6 bytes as 3 little endian ints, then bit shift
    # them together like so:    compressed_quat = w2 + (w1<<16) + (w0<<32)
    UInt16("w0"),
    UInt16("w1"),
    UInt16("w2"),
    SIZE=6, TYPE=QStruct
    )

# Translation is not compressed, and is still an xyz float triple.
translation_def = BlockDef("translation",
    Float("x"),
    Float("y"),
    Float("z"),
    )


def compressed_stream_size(node=None, parent=None, attr_index=None,
                           rawdata=None, new_value=None, *args, **kwargs):
    if attr_index is None:
        raise IndexError
    elif new_value is not None:
        return
    elif parent is None:
        raise TypeError
    elif node is None:
        node = parent[attr_index]

    if node is not None:
        return len(node) * node.itemsize
    elif rawdata is None:
        return 0

    comp_frame = parent.parent
    transform_index = comp_frame.index_by_id(parent) - 3

    offset = 44
    if transform_index != 0 or attr_index != 0:
        offset = comp_frame[transform_index][attr_index]

    # looking for the NEXT streams offset
    attr_index += 1

    if attr_index <= 3:
        end = comp_frame[transform_index][attr_index]
    elif transform_index == 2:
        end = len(rawdata)
    else:
        end = comp_frame[transform_index + 1][0]

    return end - offset


compressed_frames_def = BlockDef("compressed frames",
    Struct("rotation_offsets",
        # keyframe_head offset for rotation is directly
        # after the header, so it is ALWAYS 44
        Void("keyframe_head"),
        UInt32("keyframes"),
        UInt32("default_data"),
        UInt32("keyframe_data"),
        SIZE=12
        ),
    QStruct("translation_offsets",
        UInt32("keyframe_head"),
        UInt32("keyframes"),
        UInt32("default_data"),
        UInt32("keyframe_data"),
        SIZE=16
        ),
    QStruct("scale_offsets",
        UInt32("keyframe_head"),
        UInt32("keyframes"),
        UInt32("default_data"),
        UInt32("keyframe_data"),
        SIZE=16
        ),

    # keyframe_head:
    # Contains information on which keyframes are stored for each node
    # and where in the keyframes array to read the keyframe index from.
    # Look at the above keyframe_header struct for exact specifications.

    # keyframes:
    # An array of arrays of UInt16s. The outer array stores one array
    # of UInt16s for each node being rotated, translated, or scaled.
    # Each of these UInt16 arrays stores the keyframe numbers of which
    # frames have been stored in keyframe_data

    # default_data:
    # An array of default rotation and translation data for each node.
    # The array has one entry for each node in the animation, and the data
    # in it is given by either quaternion48_def or a translation_def.
    # Scale is ONLY stored here if the nodes scale is animated.
    # Assume a scale of 1.0 otherwise.

    # keyframe_data:
    # Contains the data for each node for each stored keyframe.
    # The data is different depending on the transformation type.
    #     (follows the same rules as default_data)

    Container("rotation",
        UInt32Array("keyframe_head", SIZE=compressed_stream_size),
        UInt16Array("keyframes", SIZE=compressed_stream_size),
        UInt16Array("default_data", SIZE=compressed_stream_size),
        UInt16Array("keyframe_data", SIZE=compressed_stream_size),
        ),
    Container("translation",
        UInt32Array("keyframe_head", SIZE=compressed_stream_size),
        UInt16Array("keyframes", SIZE=compressed_stream_size),
        FloatArray("default_data", SIZE=compressed_stream_size),
        FloatArray("keyframe_data", SIZE=compressed_stream_size),
        ),
    Container("scale",
        UInt32Array("keyframe_head", SIZE=compressed_stream_size),
        UInt16Array("keyframes", SIZE=compressed_stream_size),
        FloatArray("default_data", SIZE=compressed_stream_size),
        FloatArray("keyframe_data", SIZE=compressed_stream_size),
        ),
    endian='<'
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.shdr import *
from ..common_descs import *
from supyr_struct.defs.tag_def import TagDef

shdr_attrs = dict(shdr_attrs)
shdr_attrs[6] = SEnum16("material_type", *materials_list)

shader_body = Struct("tagdata",
    shdr_attrs,
    SIZE=40
    )

def get():
    return shdr_def

shdr_def = TagDef("shdr",
    blam_header_stubbs('shdr'),
    shader_body,

    ext=".shader", endian=">", tag_cls=ShdrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from math import sin, cos

from reclaimer.hek.defs.objs.tag import HekTag

class LighTag(HekTag):

    def calc_internal_data(self):
        HekTag.calc_internal_data(self)

        shape = self.data.tagdata.shape
        shape.cosine_falloff_angle = cos(shape.falloff_angle)
        shape.cosine_cutoff_angle  = cos(shape.cutoff_angle)
        shape.sine_cutoff_angle    = sin(shape.cutoff_angle)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.common_descs import *
from reclaimer.shadowrun_prototype.constants import *


def sr_tag_class(*args, **kwargs):
    '''
    A macro for creating a tag_class enum desc with the
    enumerations set to the provided tag_class fcc's.
    '''
    kwargs["class_mapping"] = sr_tag_class_fcc_to_ext
    return tag_class(*args, **kwargs)


def dependency(name='tag_ref', valid_ids=None, **kwargs):
    '''This function serves to macro the creation of a tag dependency'''
    if isinstance(valid_ids, tuple):
        valid_ids = sr_tag_class(*valid_ids)
    elif isinstance(valid_ids, str):
        valid_ids = sr_tag_class(valid_ids)
    elif valid_ids is None:
        valid_ids = valid_tags

    return TagRef(name,
        valid_ids,
        INCLUDE=tag_ref_struct,
        STEPTREE=StrTagRef(
            "filepath", SIZE=tag_ref_str_size, GUI_NAME="", MAX=234),
        **kwargs
        )


def blam_header(tagid, version=1):
    '''This function serves to macro the creation of a tag header'''
    header_desc = dict(tag_header)
    header_desc[1] = dict(header_desc[1])
    header_desc[5] = dict(header_desc[5])
    header_desc[1][DEFAULT] = tagid
    header_desc[5][DEFAULT] = version
    return header_desc


sr_valid_tags = sr_tag_class(*sr_tag_class_fcc_to_ext.keys())


# Descriptors
sr_tag_header = Struct("blam_header",
    Pad(36),
    UEnum32("tag_class",
        GUI_NAME="tag_class", INCLUDE=sr_valid_tags, EDITABLE=False
        ),
    UInt32("checksum", DEFAULT=0, EDITABLE=False),
    UInt32("header_size",  DEFAULT=64, EDITABLE=False),
    BBool64("flags",
        "edited_with_mozz",
        EDITABLE=False
        ),
    UInt16("version", DEFAULT=1, EDITABLE=False),
    UInt8("integrity0", DEFAULT=0, EDITABLE=False),
    UInt8("integrity1", DEFAULT=255, EDITABLE=False),
    UEnum32("engine_id",
        ("shadowrun_xbox", 'sr_x'),
        DEFAULT='sr_x', EDITABLE=False
        ),
    VISIBLE=VISIBILITY_METADATA, SIZE=64, ENDIAN=">"  # KEEP THE ENDIAN SPECIFICATION
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.animation import animation_decompilation
from reclaimer.bitmaps import bitmap_decompilation
from reclaimer.halo_script import hsc_decompilation
from reclaimer.model import model_decompilation
from reclaimer.physics import physics_decompilation
from reclaimer.sounds import sound_decompilation
from reclaimer.strings import strings_decompilation


h1_data_extractors = {
    'phys': physics_decompilation.extract_physics,
    'mode': model_decompilation.extract_model,
    'mod2': model_decompilation.extract_model,
    'antr': animation_decompilation.extract_model_animations,
    'magy': animation_decompilation.extract_model_animations,
    #'coll': extract_collision,
    #'sbsp': None, 'font': None, 'unic': None,
    'str#': strings_decompilation.extract_string_list,
    'ustr': strings_decompilation.extract_unicode_string_list,
    "hmt ": strings_decompilation.extract_hud_message_text,
    "bitm": bitmap_decompilation.extract_bitmaps,
    "snd!": sound_decompilation.extract_h1_sounds,
    'scnr': hsc_decompilation.extract_h1_scripts,
    }

h2_data_extractors = {
    #'mode', 'coll', 'phmo', 'jmad',
    #'sbsp',

    #'unic',
    "bitm": bitmap_decompilation.extract_bitmaps,
    "snd!": sound_decompilation.extract_h2_sounds,
    }

h3_data_extractors = {
    "bitm": bitmap_decompilation.extract_bitmaps,
    }



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.field_types import *

H3TagRef      = FieldType(base=TagRef, name="H3TagRef")
H3Reflexive   = FieldType(base=Reflexive,   name="H3Reflexive")
H3RawdataRef  = FieldType(base=RawdataRef,  name="H3RawdataRef")



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.ant_ import Ant_Tag
from supyr_struct.defs.tag_def import TagDef

vertex = Struct("vertex",
    Float("spring_strength_coefficient"),

    Pad(24),
    yp_float_rad("angles"),  # radians
    float_wu("length"),
    SInt16("sequence_index"),

    Pad(2),
    QStruct("color", INCLUDE=argb_float),
    QStruct("lod_color", INCLUDE=argb_float),
    Pad(40),
    QStruct('offset', INCLUDE=xyz_float, VISIBLE=False),

    SIZE=128
    )

ant__body = Struct("tagdata",
    ascii_str32("attachment_marker_name"),
    dependency("bitmaps", "bitm"),
    dependency("physics", "pphy"),

    Pad(80),
    Float("spring_strength_coefficient"),
    Float("falloff_pixels"),
    Float("cutoff_pixels"),
    Float("length"),

    Pad(36),
    reflexive("vertices", vertex, 20),
    SIZE=208
    )


def get():
    return ant__def

ant__def = TagDef("ant!",
    blam_header('ant!'),
    ant__body,

    ext=".antenna", endian=">", tag_cls=Ant_Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

# There is only 8 spaces allocated for these in the object's gamestate.
# If you have more and it updates the state for those it would write into
# unrelated fields.
HALO_1_MAX_REGIONS = 8

HALO_1_NAME_MAX_LEN = 31

HALO_1_MAX_GEOMETRIES_PER_MODEL = 256

HALO_1_MAX_MATERIALS = 256

HALO_1_MAX_NODES = 64

HALO_1_MAX_MARKERS = 256


# If a jms file is prefixed with this token it
# cannot be randomly chosen as a permutation
JMS_PERM_CANNOT_BE_RANDOMLY_CHOSEN_TOKEN = "~"

SCALE_INTERNAL_TO_JMS = 100.0

JMS_VERSION_HALO_1 = "8200"
JMS_VERSION_HALO_2_8210 = "8210"
